"""
Test implementation of something that may or may not behave like a battery
"""

from __future__ import annotations

import logging
import random

from moat.ems.battery._base import BaseBalancer, BaseBattery
from moat.lib.micro import sleep_ms

logger = logging.getLogger(__name__)


class Batt(BaseBattery):
    """
    A fake battery.

    Additional Config::
        n: 4  # number of cells
        cfg: {}  # add
        rnd: 0.1  # random factor for add_p, here 0.9…1.1


    """

    u_d: float = 1.01
    i: float = 0

    def __init__(self, cfg):
        super().__init__(cfg)
        self._rand = random.Random()

    def _random(self, r=1):
        "random number between 0 and @r"
        return self._rand.random() * r

    doc_u = dict(_d="read u", _r="float:voltage")

    async def cmd_u(self):
        "return synthetic voltage, multiplied by u_d"
        return self.u_d * await super().cmd_u()

    doc_u_d = dict(_d="write u delta", ud="float:voltage delta")

    async def cmd_u_d(self, *, ud: float):
        "change delta of battery vs. sum-of-cells voltage"
        self.u_d = ud

    doc_c = dict(_d="read charge state", _r="float")

    async def cmd_c(self):
        r = 0
        for c in self.apps:
            r += await c.cmd_c()
        return r / self.n

    doc_i = dict(_d="read current", _r="float")

    async def cmd_i(self, i: float | None = None):
        if i is not None:
            self.i = i
        return self.i

    async def feed_energy(self):
        s = 100  # ms per loop
        while True:
            await sleep_ms(s)
            p = (await self.cmd_u()) * (await self.cmd_i())

            r = self.cfg.get("rnd", 0)
            u = await self.cmd_u()
            for c in self.apps:
                f = (await c.cmd_u()) / u
                rnd = 1 + self._random(r * 2) - r if r > 0 else 1
                await c.cmd_add_p(p=p * f * rnd, t=s)

    async def start_tasks(self, tg):
        await super().start_tasks(tg)
        await tg.spawn(self.feed_energy)


class Bal(BaseBalancer):
    """
    Balancer support for a battery.
    """

    pass



"""
Error handler and retry app.
"""

from __future__ import annotations

from moat.lib.micro import L, log, sleep_ms
from moat.lib.rpc import BaseFwdCmd

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import Any


class RetryCmd(BaseFwdCmd):
    """
    An error handler and possibly-retrying subcommand manager.

    This handler catches some retryable exceptions, thus shielding the rest
    of MoaT from them.

    Parameters:
        cfg(dict): sub-App to run
        retry(int): Number of retries
        timeout(int): delay before restarting
        notify(Path): Called on error
        always(bool): also restart the app if it ends without error

    If *retry* is zero, the exception is simply ignored. Otherwise the
    app is restarted after a timeout.

    Set *retry* to -1 for infinite retries.

    Set @always to `True` if the app should be restarted if it ends without
    raising an error.

    TODO: exponential back-off
    """

    doc = dict(
        _c=dict(
            _d="Error catcher",
            cfg="app:sub-app to run",
            retry="int:autorestart# (0, -1=inf)",
            timeout="int:restart timer (100)",
            notify="path:call on error",
            always="bool:restart if no error",
        )
    )

    retry: int = 0
    timeout: int = 100
    notify: Callable[..., Awaitable[Any]] | None = None

    @property
    def cfg_name(self) -> str:
        "Human readable name"

        return (
            f"{super().cfg_name} → {self.app.cfg_name if self.app else self.cfg.get('app', '?')}"
        )

    async def handle(self, *a, **k):
        """Handle a command, waiting for ready if needed."""
        if L:
            app = self.app
            if app is None:
                await super().wait_ready()
                app = self.app
            if app is None:
                raise RuntimeError("No app")
            await app.wait_ready()
        return await super().handle(*a, **k)

    async def reload(self):
        """Reload configuration."""
        self._load()
        await super().reload()

    def _load(self):
        self.retry = self.cfg.get("retry", 0)
        self.timeout = self.cfg.get("timeout", 100)
        self.always = self.cfg.get("always", False)

        p = self.cfg.get("notify", None)
        if p is None:
            self.notify = None
        else:
            root = self.root
            if root is None:
                raise RuntimeError("Not attached")
            notify = root.sub_at(p, cmd=True)
            if not callable(notify):
                raise TypeError(f"Notify target is not callable: {p}")
            self.notify = notify

    if L:

        async def wait_ready(self, wait: bool = True):
            """Allow for non-restarted sub-app."""
            while res := await super().wait_ready(wait=wait):
                if not self.retry:
                    return res
                await sleep_ms(1)

            return res

    async def run_app(self):
        """
        Runs the sub-app and handles restarting and error shielding.
        """
        self._load()

        while True:
            try:
                # log("Err Run %s %r", self.path, self)
                await super().run_app()
            except Exception as exc:
                log("Err %s %r", self.path, exc, err=exc)
                if self.notify is not None:
                    try:
                        await self.notify(here=self.path, err=exc)
                    except Exception:
                        log("Err Report %s %r", self.path, exc, err=exc)

                if not self.retry:
                    if self.cfg.get("retry", 0):
                        raise
                    return
            else:
                # ends without error
                log("Err End %s %r", self.path, self.app)
                if not self.always:
                    return

            if self.retry:
                app = self.app
                if app is None:
                    raise RuntimeError("No app")
                app.init_events()
            # otherwise dead

            if self.retry > 0:
                self.retry -= 1
            await sleep_ms(self.timeout)



"""
Codec library.
"""

from __future__ import annotations

from ._base import Codec, Extension, NoCodecError

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._base import ByteType, VarByteType  # noqa:F401


__all__ = ["Codec", "Extension", "NoCodecError", "get_codec"]


def get_codec(codec: str | Codec, **kw) -> Codec:
    """
    Loads and initializes the named codec.
    """
    from moat.util import import_  # noqa: PLC0415

    if isinstance(codec, Codec):
        return codec
    if codec is None:
        codec = "noop"
    if "." not in codec:
        if codec.startswith("std-"):
            codec = "moat.lib.codec.moat_" + codec[4:]
        else:
            codec = "moat.lib.codec." + codec
    return import_(codec).Codec(**kw)



"""
The main entry point to load MoaT.
"""

from __future__ import annotations

import gc as _gc
import os as _os

_gc.collect()
_fm = _gc.mem_free()
_fa = _gc.mem_alloc()


def go(state=None, cmd=True):
    """
    Start MoaT.

    The system state can be passed as an argument; otherwise it's read from
    RTC, Flash, or defaults to "skip".

    If @cmd is True, the system has been started manually. This prevents
    processing of error tracebacks. Also, a fake "OK" prompt is emitted, to
    fool the host's command line processing.

    The state can be a comma-separated list: the next state in the list
    will be used on reboot.

    * skip
      Do nothing: exit to MicroPython prompt.

    * std
      Work normally, i.e. load from /lib, /rom/lib, .frozen.
      mode if there's a problem. (Warning: Switching to fallback doesn't
      always work, esp. when you run out of memory or otherwise hard-crash
      the system. Always set up a hardware watchdog, if available.)

    * rom
      As above, but don't use /lib. Uses "moat_rom.cfg" if available.

    * flash
      Only use .frozen. Uses "moat_flash.cfg" if available.

    * norom
      Uses /lib and .frozen. Uses "moat_nr.cfg" if available.

    The default is "fallback" if moat_fb.cfg exists, else "rom" if
    moat_rom.cfg exists, else whatever is in "moat.cfg" ('mode' key)
    """

    # copy from moat.util.compat
    import sys as sys  # noqa: PLC0415
    import time  # noqa: PLC0415

    from moat.micro.rtc import RTC  # noqa: PLC0415

    RTC.init({
        "use": ["mem", {"name": "fs", "dest": "moat.data", "direct": {"state": "moat.state"}}]
    })

    import machine  # noqa: PLC0415

    def log(s, *x, err=None):
        if x:
            s = s % x
        print(s, file=sys.stderr)
        if err is not None:
            sys.print_exception(err, sys.stderr)

    states = [
        ("flash", "moat_fb.cfg"),
        ("rom", "moat_rom.cfg"),
        ("safe", "moat_safe.cfg"),
        ("std", "moat.cfg"),
    ]
    if state is None:
        state = RTC.get_sync("state", default=None)
    if state is None:
        for st, fn in states:
            try:
                _os.stat(fn)
            except OSError:
                pass
            else:
                state = st
                break
    if state is None:
        state = "once"
    if state == "once":
        new_state = "skip"
    else:
        try:
            state, new_state = state.split(",", 1)
        except ValueError:
            new_state = state

    if state != new_state:
        RTC.set_sync("state", new_state, fs=None, all=False)
    if state == "skip":
        log("MoaT state: %r", state)
        return True

    fn = dict(states).get(state, "moat.cfg")

    # clean up default stuff from path
    for p in ("", "/", ".", "/lib", ".frozen", "/rom", "/rom/lib"):
        try:
            sys.path.remove(p)
        except ValueError:
            pass

    # build path
    lin = sys.platform == "linux"
    print("*** STATE ***", state, "***", file=sys.stderr)
    if state in ("norom", "std", "once", "safe"):
        sys.path.append("lib" if lin else "/lib")
    if state in ("rom", "std"):
        sys.path.append("/rom")
    if state in ("flash", "rom", "norom", "std", "once", "safe"):
        sys.path.append(".frozen")
    # keep the root in the path, but at the end
    sys.path.append("." if lin else "/")

    print("Start MoaT:", repr(state), file=sys.stderr)

    from moat.micro.main import main, wr_exc  # noqa: PLC0415
    from moat.micro.rtc import at  # noqa: PLC0415

    i = dict(cfg=fn, s=state, ns=new_state, fm=_fm, fa=_fa, fb=state != "std")

    if cmd:
        at("main2", i)
        try:
            main(fn, i=i, fake_end=False)
        except ImportError:
            print("PATH:", sys.path, file=sys.stderr)
            raise
        return

    try:
        at("main1", i)

        from moat.util import ungroup  # noqa:PLC0415

        with ungroup:
            main(fn, i=i)

    except ImportError as exc:
        print("PATH:", sys.path, file=sys.stderr)
        wr_exc(exc)
        raise

    except KeyboardInterrupt:
        at("main3", i)
        print("MoaT stopped, mode is 'skip'", file=sys.stderr)
        # because the watchdog might kill us
        RTC.set_sync("state", "skip", all=True)

    except SystemExit as exc:
        at("main4", i)
        print("REBOOT to", new_state, file=sys.stderr)
        time.sleep_ms(100)
        if exc.val:
            machine.reset()
        else:
            machine.soft_reset()

    except BaseException as exc:
        wr_exc(exc)
        at("main5", i)
        del main
        sys.modules.clear()

        if sys.platform == "linux":
            if isinstance(exc, EOFError):
                log("MoaT stopped: EOF")
                sys.exit(0)
            log("CRASH! Exiting!", err=exc)
            sys.exit(1)

        RTC.set_sync("state", new_state, all=False)

        log("CRASH! %r :: REBOOT to %r", exc, new_state, err=exc)
        time.sleep_ms(1000)
        machine.soft_reset()

    else:
        at("main6", i)
        try:
            from moat.micro.console import main as m  # noqa:PLC0415

            if m.console is not None:
                return
        except Exception:  # noqa:S110
            pass
        log("MoaT Ended.")
        sys.exit(0)


def g():
    "shortcut for ``go('once')``"
    go("once")



# ruff:noqa:F405
"""
MoaT's command multiplexer.
"""

from __future__ import annotations

from moat import DOC as _DOC

# Import BaseMsgHandler at runtime for Sphinx documentation
from .base import BaseMsgHandler as BaseMsgHandler

# Import constants directly (not lazy)
from .const import *  # noqa: F403

# Import errors directly
from .errors import *  # noqa: F403

from typing import TYPE_CHECKING as _TC

# Lazy loading for classes and functions
_imports = {
    # From anyio
    "rpc_on_aiostream": "anyio",
    # From base
    "BaseMsgHandler": "base",
    "Caller": "base",
    "Key": "base",
    "MsgHandler": "base",
    "MsgSender": "base",
    "OptDict": "base",
    "SubMsgSender": "base",
    # From cmd.array
    "ArrayCmd": "cmd.array",
    # From cmd.retry
    "RetryCmd": "cmd.retry",
    # From cmd._base and cmd.base
    "APP": "cmd._base",  # not exported/documented
    "add_app_prefix": "cmd._base",
    "load_app": "cmd._base",
    "BaseCmd": "cmd._base",
    "LoadCmd": "cmd._base",
    "LockBaseCmd": "cmd._base",
    "RootCmd": "cmd.base",
    # From cmd.tree.dir
    "BaseSubCmd": "cmd.tree.dir",
    "BaseSuperCmd": "cmd.tree.dir",
    "CfgStore": "cmd.tree.dir",
    "DirCmd": "cmd.tree.dir",
    "SubStore": "cmd.tree.dir",
    # From cmd.tree.layer
    "BaseFwdCmd": "cmd.tree.layer",
    "BaseLayerCmd": "cmd.tree.layer",
    # From cmd.tree.listen
    "BaseListenCmd": "cmd.tree.listen",
    "BaseListenOneCmd": "cmd.tree.listen",
    # From conn
    "BaseConnIter": "conn.util",
    "TcpIter": "conn.tcp",
    "UnixIter": "conn.unix",
    "WsIter": "conn.ws",
    # From msg
    "Msg": "msg",
    "MsgLink": "msg",
    "MsgResult": "msg",
    # From nest
    "rpc_on_rpc": "nest",
    # From stream.base
    "HandlerStream": "stream.base",
    "StreamLink": "stream.base",
    "i_f2wire": "stream.base",
    "wire2i_f": "stream.base",
    # From stream.cmdbbm
    "BaseCmdBBM": "stream.cmdbbm",
    # From cmd.msg
    "MsgStream": "cmd.msg",
    "BaseCmdMsg": "cmd.msg",
    "CmdMsg": "cmd.msg",
    "SingleCmdMsg": "cmd.msg",
    "ExtCmdMsg": "cmd.msg",
    # From stream.xcmd
    "BBMCmd": "stream.xcmd",
    "MsgCmd": "stream.xcmd",
    "BufCmd": "stream.xcmd",
    "BlkCmd": "stream.xcmd",
    # From log
    "Logger": "cmd.log",
    # From auth
    "Auth": "auth._base",
    # From alert
    "Alert": "alert",
    "AlertHandler": "alert",
}


def __getattr__(attr: str):
    try:
        mod = _imports[attr]
    except KeyError:
        raise AttributeError(attr) from None
    value = getattr(__import__(mod, globals(), None, (attr,), 1), attr)
    globals()[attr] = value
    return value


__all__ = [  # noqa:RUF022
    # From const (not lazy)
    "B_STREAM",
    "B_ERROR",
    "B_WARNING",
    "B_WARNING_INTERNAL",
    "B_FLAGSTR",
    "E_UNSPEC",
    "E_NO_STREAM",
    "E_CANCEL",
    "E_NO_CMDS",
    "E_SKIP",
    "E_MUST_STREAM",
    "E_ERROR",
    "E_NO_CMD",
    "S_END",
    "S_NEW",
    "S_ON",
    "S_OFF",
    "SD_NONE",
    "SD_IN",
    "SD_OUT",
    "SD_BOTH",
    # From errors
    "NotReadyError",
    "ShortCommandError",
    "LongCommandError",
    "RemoteError",
    "StreamError",
    "Flow",
    "StopMe",
    "SkippedData",
    "NoStream",
    "NoCmds",
    "NoCmd",
    "WantsStream",
    "MustStream",
    # From anyio
    "AioStream",
    "rpc_on_aiostream",
    # From base
    "BaseMsgHandler",
    "Caller",
    "Key",
    "MsgHandler",
    "MsgSender",
    "OptDict",
    "SubMsgSender",
    # From cmd.array
    "ArrayCmd",
    # From cmd.retry
    "RetryCmd",
    # From cmd.base
    "BaseCmd",
    "add_app_prefix",
    "load_app",
    "LoadCmd",
    "LockBaseCmd",
    "RootCmd",
    # From cmd.tree.dir
    "BaseSubCmd",
    "BaseSuperCmd",
    "CfgStore",
    "DirCmd",
    "SubStore",
    # From cmd.tree.layer
    "BaseFwdCmd",
    "BaseLayerCmd",
    # From cmd.tree.listen
    "BaseListenCmd",
    "BaseListenOneCmd",
    # From conn
    "BaseConnIter",
    "TcpIter",
    "UnixIter",
    "WsIter",
    # From msg
    "Msg",
    "MsgLink",
    "MsgResult",
    # From nest
    "CmdStream",
    "rpc_on_rpc",
    # From stream.base
    "HandlerStream",
    "StreamLink",
    "i_f2wire",
    "wire2i_f",
    # From stream.cmdbbm
    "BaseCmdBBM",
    # From cmd.msg
    "MsgStream",
    "BaseCmdMsg",
    "CmdMsg",
    "SingleCmdMsg",
    "ExtCmdMsg",
    # From stream.xcmd
    "BBMCmd",
    "MsgCmd",
    "BufCmd",
    "BlkCmd",
    # From log
    "Logger",
    # From auth
    "Auth",
    # From alert
    "Alert",
    "AlertHandler",
]

if _TC or _DOC:
    _DOC = False
    from .anyio import AioStream as AioStream  # noqa:I001
    from .base import Caller as Caller
    from .base import Key as Key
    from .base import MsgHandler as MsgHandler
    from .base import MsgSender as MsgSender
    from .base import OptDict as OptDict
    from .base import SubMsgSender as SubMsgSender
    from .cmd._base import BaseCmd as BaseCmd
    from .cmd._base import add_app_prefix as add_app_prefix
    from .cmd._base import load_app as load_app
    from .cmd._base import LoadCmd as LoadCmd
    from .cmd._base import LockBaseCmd as LockBaseCmd
    from .cmd.array import ArrayCmd as ArrayCmd
    from .cmd.retry import RetryCmd as RetryCmd
    from .cmd.base import RootCmd as RootCmd
    from .cmd.log import Logger as Logger
    from .cmd.tree.dir import BaseSubCmd, BaseSuperCmd, CfgStore, SubStore
    from .cmd.tree.layer import BaseFwdCmd, BaseLayerCmd
    from .cmd.tree.listen import BaseListenCmd, BaseListenOneCmd
    from .conn.tcp import TcpIter as TcpIter
    from .conn.unix import UnixIter as UnixIter
    from .conn.util import BaseConnIter as BaseConnIter
    from .conn.ws import WsIter as WsIter
    from .msg import Msg as Msg
    from .msg import MsgLink as MsgLink
    from .msg import MsgResult as MsgResult
    from .stream.base import HandlerStream as HandlerStream
    from .stream.base import StreamLink as StreamLink
    from .stream.base import i_f2wire as i_f2wire
    from .stream.base import wire2i_f as wire2i_f
    from .stream.cmdbbm import BaseCmdBBM as BaseCmdBBM
    from .cmd.msg import BaseCmdMsg as BaseCmdMsg
    from .cmd.msg import CmdMsg as CmdMsg
    from .cmd.msg import ExtCmdMsg as ExtCmdMsg
    from .cmd.msg import MsgStream as MsgStream
    from .cmd.msg import SingleCmdMsg as SingleCmdMsg
    from .stream.xcmd import BBMCmd as BBMCmd
    from .stream.xcmd import BlkCmd as BlkCmd
    from .stream.xcmd import BufCmd as BufCmd
    from .stream.xcmd import MsgCmd as MsgCmd
    from .nest import CmdStream as CmdStream
    from .nest import rpc_on_rpc as rpc_on_rpc
    from .alert import Alert as Alert
    from .alert import AlertHandler as AlertHandler
    from .auth._base import Auth as Auth



"""
fake sensors
"""

from __future__ import annotations

from moat.lib.rpc import BaseCmd

PINS = {}


class NoOp(BaseCmd):
    """
    This is command that does nothing.
    """

    doc = dict(_c=dict(_d="NoOp"))



"""Line buffering utility for async output."""

from __future__ import annotations

import sys
from inspect import iscoroutinefunction

from moat.lib.micro import (
    ACM,
    AC_exit,
    Event,
    TaskGroup,
    TimeoutError,  # noqa:A004
    is_async,
    wait_for_ms,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import Any, Self

__all__ = ["Liner"]


async def _no_op() -> None:
    pass


class Liner:
    """
    A way to collect output until timeout-or-EOL
    and then to print it all at once.


    Args:
        prefix:
            Text to emit at the start of each line.
        incomplete:
            String to add to incomplete lines.
        writer:
            Called to write complete lines. Must not append linefeed.
        delay:
            Milliseconds until a line is declared incomplete. Default 100.

    If the writer is async, so is calling this obj with data.
    """

    def __init__(
        self,
        prefix: str = "",
        incomplete: str = "…",
        writer: Callable[[str], Any | Awaitable[Any]] = sys.stdout.write,
        delay: int = 100,
    ) -> None:
        self.prefix: str = prefix
        self.incomplete: str = incomplete
        self.writer: Callable[[str], Any | Awaitable[Any]] = writer
        self.delay: int = delay

        self.buf: bytearray | None = bytearray()
        self.evt: Event
        self._tg: Any  # TaskGroup from moat.lib.micro

    async def __aenter__(self) -> Self:
        AC = ACM(self)
        try:
            self.evt = Event()
            self._tg = tg = await AC(TaskGroup())
            tg.start_soon(self._flush)
        except BaseException as exc:
            await AC_exit(self, type(exc), exc, None)
            raise
        return self

    async def __aexit__(self, *exc: object) -> Any:
        self._tg.cancel()
        try:
            if self.buf:
                await self._partial(True)
        except BaseException as err:
            try:
                await AC_exit(self, type(err), err, getattr(err, "__traceback__", None))
            except BaseException as ex:
                if ex is not err:
                    raise
            raise
        return await AC_exit(self, *exc)

    async def _flush(self) -> None:
        # Background task to write incomplete lines
        while True:
            try:
                await wait_for_ms(self.delay, self.evt.wait)
            except TimeoutError:
                pass
            else:
                # Event triggered
                self.evt = Event()
                continue

            # Event did not trigger
            if self.buf:
                await self._partial()

            # now wait for the next possibly-incomplete line
            await self.evt.wait()
            self.evt = Event()

    async def _partial(self, end: bool = False) -> None:
        pr = self.buf.decode("utf-8")  # type: ignore[union-attr]  # checked that buf is not None
        self.buf = None if end else bytearray()

        res = self.writer(self.prefix + pr + self.incomplete + ("-END-" if end else "") + "\n")
        if is_async(res):
            await res

    def __call__(self, data: bytes | memoryview) -> Any | Awaitable[Any]:
        """
        Writer.
        """
        if isinstance(data, memoryview):
            data = bytes(data)
        buf = self.buf
        if buf is None:
            # after end. Oops, async timing is nontrivial.
            buf = data
            idx = len(buf)
        else:
            buf += data
            idx = buf.rfind(b"\n")
        try:
            if idx != -1:
                pr = buf[:idx].decode("utf-8").replace("\n", f"\n{self.prefix}")
                self.buf = bytearray(buf[idx + 1 :])
                return self.writer(self.prefix + pr + "\n")
            elif iscoroutinefunction(self.writer):
                return _no_op()
        finally:
            if self.buf:
                self.evt.set()



"""
Remote link app for MoaT message exchange.
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc.cmd.msg import BaseCmdMsg
from moat.lib.rpc.stream.xcmd import BufCmd
from moat.lib.stream import serial_stack

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.stream import BaseMsg


class Link(BaseCmdMsg):
    """
    Connects to a `moat.lib.rpc.stream.cmdbbm.BaseCmdBBM` object
    exporting a `moat.lib.stream.BaseBuf`.
    """

    doc = dict(_c=dict(_d="Command forwarding to remote stream", path="path:dest"))

    async def stream(self) -> BaseMsg:
        """Returns the stack-wrapped link."""
        sd = BufCmd(self.cfg)
        return await AC_use(self, serial_stack(sd, self.cfg))



"""
A hacked-up copy of some parts of `moat.util`.
"""
# ruff: noqa: E402

from __future__ import annotations

from copy import deepcopy

from async_queue import Queue, QueueEmpty, QueueFull  # noqa:F401

NotGiven = Ellipsis

from moat.lib.micro import Event, log
from moat.lib.path import Path

from .exc import ExpAttrError as ExpAttrError
from .exc import ExpectedError as ExpectedError
from .exc import ExpKeyError as ExpKeyError
from .exc import ungroup as ungroup
from .misc import de_async as de_async
from .misc import wait_complain as wait_complain
from .part import enc_part as enc_part
from .part import get_part as get_part
from .part import set_part as set_part
from .pp import pop_kw as pop_kw
from .pp import push_kw as push_kw

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable


class OutOfData(EOFError):  # noqa: D101
    pass


class CancelledError(Exception):
    """
    Not an asyncio-style cancellation
    """


class OptCtx:
    "optional context"

    def __init__(self, obj=None):
        self.obj = obj

    def __enter__(self):
        if self.obj is not None:
            return self.obj.__enter__()
        return self

    def __exit__(self, *tb):
        if self.obj is not None:
            return self.obj.__exit__(*tb)


class ValueEvent:
    """A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event`.

    An event object manages an internal value, which is initially
    unset, and a task can wait for it to become True.

    Args:
      ``scope``:  A cancelation scope that will be cancelled if/when
                  this ValueEvent is. Used for clean cancel propagation.

    Note that the value can only be read once.
    """

    event = None
    value = None

    def __init__(self, scope=None):
        self.event = Event()
        self.scope = scope

    def set(self, value):
        """Set the result to return this value, and wake any waiting task."""
        self.value = value
        self.event.set()

    def set_error(self, exc):
        """Set the result to raise this exceptio, and wake any waiting task."""
        if isinstance(exc, type):
            exc = exc()
        self.value = exc
        self.event.set()

    def is_set(self):
        """Check whether the event has occurred."""
        return self.value is not None

    def cancel(self):
        """Send a cancelation to the recipient.

        TODO: Trio can't do that cleanly.
        """
        if self.scope is not None:
            self.scope.cancel()
        self.set_error(CancelledError())

    async def wait(self):
        """Block until the value is set.

        If it's already set, then this method returns immediately.

        The value is not (yet) read; if it's an error, it will not be raised from here.
        """
        await self.event.wait()

    async def get(self):
        """Block until the value is set.

        If it's already set, then this method returns immediately.

        The value can only be read once.
        """
        await self.event.wait()
        if isinstance(self.value, Exception):
            raise self.value
        return self.value


class attrdict(dict):
    """
    A dict that can be accessed via attribute syntax.

    This is a very minimal implementation. While it does supports an update
    hook, it doesn't remember having been updated.
    """

    updated_: Callable = lambda _x: None

    def __getattr__(self, k, d=NotGiven):
        try:
            return self[k]
        except KeyError:
            if d is NotGiven:
                raise AttributeError(k) from None
            return d

    def __setattr__(self, k, v):
        if k[0] == "_" or k[-1] == "_":
            object.__setattr__(self, k, v)
        else:
            self[k] = v

    def __delattr__(self, k):
        if k[0] == "_" or k[-1] == "_":
            object.__delattr__(self, k)
        else:
            try:
                del self[k]
            except KeyError:
                return AttributeError(k)


def to_attrdict(d) -> attrdict:
    """
    Return a hierarchy with all dicts converted to attrdicts.
    """
    if isinstance(d, dict):
        return attrdict((k, to_attrdict(v)) for k, v in d.items())
    if isinstance(d, Path):
        # this is not in the CPython version because there,
        # `Path` is not a subclass of `tuple`
        return d
    if isinstance(d, (tuple, list)):
        return [to_attrdict(v) for v in d]
    return d


def import_(name, off=0):
    """
    Import a module and access an object in it.

    `import_("a.b.c.d.e", 2)` imports "a.b.c" and returns the e attribute
    of object d from it.
    """
    n = name.split(".")
    mn = ".".join(n[: -off if off else 99])
    try:
        res = __import__(mn)
        for nn in n[1:]:
            res = getattr(res, nn)
    except Exception as exc:
        log("ERR loading %s: %r", name, exc)
        raise
    return res


def load_from_cfg(cfg, *a, _raise=False, **k):
    """
    A simple frontend to load a module, access a class/object from it,
    and call that with the config (and whichever other arguments you want to
    use).

    The module+object name is the "client" attribute.
    """
    if "client" not in cfg:
        if _raise:
            raise ValueError("must be configured")
        return None
    m = import_(cfg.client, off=1)
    return m(cfg, *a, **k)


class Lockstep:
    """
    A lock-step buffer.

    Very simple, but works only for one reader and one write.
    """

    def __init__(self):
        self.q = Queue(0)
        self._get = Event()
        self._put = Event()

    def __aiter__(self):
        return self

    async def __anext__(self):
        """
        Reads from the buffer.

        Waits for the writer to show up before proceeding.
        """
        self._get.set()
        await self._put.wait()
        s, self.s = self.s, None

        self._put = Event()
        return s

    get = __anext__

    async def put(self, s):
        """
        Write to the buffer.

        Waits for the reader to show up before proceeding.
        """
        await self._get.wait()
        self.s = s
        self._put.set()
        self._get = Event()


def exc_iter(exc):
    """
    iterate over all non-exceptiongroup parts of an exception(group)
    """
    from moat.lib.micro import BaseExceptionGroup, ExceptionGroup  # noqa: PLC0415,A004

    if isinstance(exc, (ExceptionGroup, BaseExceptionGroup)):
        for e in exc.exceptions:
            yield from exc_iter(e)
    else:
        yield exc


def combine_dict(*d, cls=dict, deep=False) -> dict:
    """
    Returns a dict with all keys+values of all dict arguments.
    The first found value wins.

    This operation is recursive and non-destructive. If `deep` is set, the
    result always is a deep copy.

    A value of `NotGiven` causes an entry to be skipped.

    TODO: arrays are not merged.

    Args:
      cls (type): a class to instantiate the result with. Default: dict.
        Often used: :class:`attrdict`.
      deep (bool): if set, always copy.
    """
    res = cls()
    keys = {}
    if not d:
        return res

    if len(d) == 1 and deep and not isinstance(d[0], dict):
        if deep and isinstance(d[0], (list, tuple)):
            return deepcopy(d[0])
        else:
            return d[0]

    for kv in d:
        if kv is None:
            continue
        for k, v in kv.items():
            if k not in keys:
                keys[k] = []
            keys[k].append(v)

    for k, v in keys.items():
        if v[0] is NotGiven:
            pass
        elif len(v) == 1 and not deep:
            res[k] = v[0]
        elif not isinstance(v[0], dict):
            for vv in v[1:]:
                assert vv is NotGiven or not isinstance(vv, dict)
            if deep and isinstance(v[0], (list, tuple)):
                res[k] = deepcopy(v[0])
            else:
                res[k] = v[0]
        else:
            res[k] = combine_dict(*v, cls=cls)

    return res


# Merge.


def _merge_dict(d, other, drop=False, replace=True):
    for key, value in other.items():
        if value is NotGiven:
            d.pop(key, None)
        elif key in d:
            d[key] = _merge_one(d[key], value, drop=drop, replace=replace)
        else:
            d[key] = value

    if drop:
        keys = []
        for k in d:
            if k not in other:
                keys.append(k)
        for k in keys:
            del d[k]


def _merge_list(item, value, drop=False, replace=True):
    off = 0
    if isinstance(value, (list, tuple)):
        # two lists
        lim = len(item)
        for i in range(min(lim, len(value))):
            if value[i] is NotGiven:
                item.pop(i - off)
                off += 1
            else:
                item[i - off] = _merge_one(item[i - off], value[i], drop=drop, replace=replace)

        while len(item) + off < len(value):
            val = value[len(item) + off]
            if val is NotGiven:
                off += 1
            else:
                item.append(val)

        if drop:
            while len(item) + off > len(value):
                item.pop()

    else:
        # a list, and a dict with updated/new/deleted values
        for i in range(len(item)):
            if i in value:
                val = value[i]
                if val is NotGiven:
                    item.pop(i - off)
                    off += 1
                else:
                    item[i - off] = _merge_one(item[i - off], val, drop=drop, replace=replace)

        while len(item) + off in value:
            val = value[len(item) + off]
            if val is NotGiven:
                off += 1
            else:
                item.append(val)


def _merge_one(d, other, drop=False, replace=True):
    if isinstance(d, dict):
        if isinstance(other, dict):
            _merge_dict(d, other, drop=drop, replace=replace)
        else:
            return other if replace else d
    elif isinstance(d, list):
        if isinstance(other, (dict, list, tuple)):
            _merge_list(d, other, drop=drop, replace=replace)
        else:
            return other if replace else d
    else:
        if replace:
            return d if other is None else other
        else:
            return other if d is None else d
    return d


def merge(d, *others, drop=False, replace=True):
    """
    Deep-merge a "source" and one or more "replacement" data structures.

    In contrast to `combine_dict`, the source is modified in-place.

    Rules:
    * two dicts: recurse into same-key items, and adds new items
    * two lists: recurse into same-index items
      append to the source if the replacement is longer
    * a list and a dict treats the dict as a sparse list. The value of numeric keys just
      beyond the list's length are appended, others are ignored.
    * a replacement value of `NotGiven` deletes the source entry
    * otherwise, use the second argument if it is not None, otherwise the first
      * except that if "replace" is False, these are swapped

    If "drop" is set, delete source keys that are not in the destination.
    This is useful for in-place replacements.
    """
    for other in others:
        d = _merge_one(d, other, drop=drop, replace=replace)
    return d


def _add_obj(a, b):
    """add attributes of B to A if they're missing"""
    for k in dir(b):
        if not hasattr(a, k):
            setattr(a, k, getattr(b, k))



"""
Adapter for MicroPython serial ports.
"""

from __future__ import annotations

import machine as M

from moat.util import attrdict, import_
from moat.lib.micro import AC_use, TimeoutError, log, sleep, wait_for_ms  # noqa:A004
from moat.lib.rpc.conn.file import FileBuf


# Serial link driver
# cfg:
# uart: N
# tx: PIN
# rx: PIN
# baud: 9600
#
class NamedSerial(FileBuf):
    """
    Interface to a MicroPython serial port that's already open,
    via a module name.
    """

    doc = dict(_c=dict(_d="Serial port", port="int:port# (1)", timeout="int:timeout(ms,50)"))

    def __init__(self, cfg):
        super().__init__(cfg=cfg, timeout=cfg.get("timeout", 50))

    async def stream(self):  # noqa: D102
        return import_(self.cfg["port"], 1)

    async def wr(self, buf):  # noqa: D102
        if len(buf) == 64:
            buf = memoryview(buf)
            await super().wr(buf[:32])
            await super().wr(buf[32:])
            return 64
        else:
            return await super().wr(buf)


class USBSerial(FileBuf):
    """
    Interface to a MicroPython serial port that's already open,
    via a module name.
    """

    async def stream(self):  # noqa: D102
        import moat  # noqa: PLC0415

        return moat.SERIAL


class Serial(NamedSerial):
    """
    Interface to a MicroPython serial port.
    """

    # inherits from NamedSerial for __init__ which is the same

    pack = None

    doc = dict(
        _c=dict(
            _d="Serial port",
            port="int:port# (1)",
            timeout="int:timeout(ms,50)",
            tx="int:TX pin (out)",
            rx="int:RX pin (in)",
            rts="int:RTS pin (out)",
            cts="int:CTS pin (in)",
            dtr="int:DTR pin (out)",
            txb="int:TX buffer (128)",
            rxb="int:RX buffer (128)",
            baudrate="int:rate",
            parity="bool|None: odd parity?",
            bits="int: size (8)",
            flow="str: RC flow control (no)",
            rts_state="bool:RTS on?",
            dtr_state="bool:DTR on?",
            rts_flip="bool:RTS flip on open?",
            dtr_flip="bool:DTR flip on open?",
            delay="float:wait after open",
            dtr_rts="float:wait between dtr and rts flip",
            delay_flip="float:wait after flip",
            flush="float|bool:flush inbuf after open (ms, True=200)",
        )
    )

    async def stream(self):
        "opens the port, does flushing and RTS/CTS"
        pin_rts = pin_cts = pin_dtr = None
        cfg = self.cfg
        uart_cfg = {}
        p = cfg.get("mode", attrdict())
        fl = p.get("flow", "")
        if "tx" in cfg:
            uart_cfg["tx"] = M.Pin(cfg["tx"], M.Pin.OUT)
        if "rx" in cfg:
            uart_cfg["rx"] = M.Pin(cfg["rx"])
        if "rts" in cfg:
            pin_rts = M.Pin(cfg["rts"], M.Pin.OUT)
            if "R" in fl:
                uart_cfg["rts"] = pin_rts
        if "cts" in cfg:
            pin_cts = M.Pin(cfg["cts"])
            if "C" in fl:
                uart_cfg["cts"] = pin_cts
        if "dtr" in cfg:
            pin_dtr = M.Pin(cfg["dtr"], M.Pin.OUT)
        uart_cfg["txbuf"] = cfg.get("txb", 128)
        uart_cfg["rxbuf"] = cfg.get("rxb", 128)

        uart_cfg["baudrate"] = p.get("rate", 9600)
        uart_cfg["parity"] = p.get("parity", None)
        uart_cfg["bits"] = p.get("bits", 8)
        f = 0
        if "C" in fl:
            f |= M.UART.CTS
        if "R" in fl:
            f |= M.UART.RTS
        uart_cfg["stop"] = p.get("stop", None) or 1  # no 1.5 stop bits

        rts = p.get("rts_state", 1)
        dtr = p.get("dtr_state", 1)
        dtr_rts = p.get("dtr_rts", 0)
        rts_flip = p.get("rts_flip", 0)
        dtr_flip = p.get("dtr_flip", 0)
        delay = p.get("delay", 0)
        delay_flip = p.get("delay_flip", 0.2)

        await sleep(delay)
        if rts_flip or dtr_flip:
            if dtr_rts >= 0:
                if pin_dtr is not None:
                    pin_dtr.value(dtr_flip ^ dtr)
                await sleep(dtr_rts)
                if pin_rts is not None:
                    pin_rts.value(rts_flip ^ rts)
            else:
                if pin_rts is not None:
                    pin_rts.value(rts_flip ^ rts)
                await sleep(-dtr_rts)
                if pin_dtr is not None:
                    pin_dtr.value(dtr_flip ^ dtr)
            await sleep(delay_flip)

        if dtr_rts > 0:
            if pin_dtr is not None:
                pin_dtr.value(dtr)
            await sleep(dtr_rts)
            if pin_rts is not None:
                pin_rts.value(rts)
        else:
            if pin_rts is not None:
                pin_rts.value(rts)
            await sleep(-dtr_rts)
            if pin_dtr is not None:
                pin_dtr.value(dtr)

        ser = M.UART(cfg.get("port", 0), **uart_cfg)
        ser.rts = rts
        ser.dtr = dtr
        await AC_use(self, ser.deinit)

        if (t := cfg.get("flush", None)) is not None:
            if t is True:
                t = 200
            while True:
                buf = bytearray(32)
                try:
                    n = await wait_for_ms(t, ser.rd, buf)
                except TimeoutError:
                    break
                else:
                    log("Flush: %r", buf[:n])

        return ser



# This build is used for mostly-interactive testing.
# We thus don't freeze MoaT here.
# include("$(VARIANT_DIR)/../../../../manifest.py")

# Do need the rest of it, though.

require("copy")
require("collections")
require("contextlib")
require("functools")
require("inspect")
require("pprint")
require("types")
require("typing")
require("__future__")

include("$(MPY_DIR)/extmod/asyncio")
include("$(PORT_DIR)/variants/manifest.py")



"""
Serial ports on Unix
"""

from __future__ import annotations

import anyio
import sys

from anyio_serial import Serial as _Serial
from serial import SerialException

from moat.util import attrdict
from moat.lib.micro import AC_use, log
from moat.lib.stream import AnyioBuf


# Serial link driver
# cfg:
# uart: N
# tx: PIN
# rx: PIN
# baud: 9600
#
class Serial(AnyioBuf):
    """
    We don't have numbered serial ports on Unix.
    """

    def __init__(self, *a, **k):
        raise NotImplementedError("Use NamedSerial on Unix")


class NamedSerial(AnyioBuf):
    """
    Serial port abstraction.
    """

    pack = None

    async def stream(self):  # noqa:D102
        cfg = self.cfg
        uart_cfg = {}
        uart_cfg["port"] = cfg["port"]

        p = cfg.get("mode", attrdict())

        uart_cfg["baudrate"] = p.get("rate", 9600)
        pa = p.get("parity", None)
        if pa:
            pa = "O"
        elif pa is not None:
            pa = "E"
        else:
            pa = "N"
        uart_cfg["parity"] = pa

        fl = p.get("flow", None)
        if fl:
            if "R" not in fl or "C" not in fl:
                if "R" in fl or "C" in fl:
                    raise ValueError("no support for partial flow control")
            uart_cfg["rtscts"] = True
        uart_cfg["stopbits"] = p.get("stop", None) or 1  # no 1.5 stop bits
        uart_cfg["bytesize"] = p.get("bits", 8)

        rts = p.get("rts_state", 1)
        dtr = p.get("dtr_state", 1)
        dtr = p.get("dtr_state", 1)
        dtr_rts = p.get("dtr_rts", 0)
        rts_flip = p.get("rts_flip", 0)
        dtr_flip = p.get("dtr_flip", 0)
        delay = p.get("delay", 0)
        delay_flip = p.get("delay_flip", 0.2)

        try:
            ser = await AC_use(self, _Serial(**uart_cfg))
        except SerialException as exc:
            print(
                f"""
*** Serial port not available ***
Port: {uart_cfg["port"]}
Error: {exc}"

Waiting """,
                end="",
                file=sys.stderr,
            )
            for _ in range(30):
                await anyio.sleep(1)
                try:
                    ser = await AC_use(self, _Serial(**uart_cfg))
                except SerialException:
                    print(".", end="", file=sys.stderr)
                    sys.stderr.flush()
                else:
                    print(" OK.")
                    break
            else:
                raise

        await anyio.sleep(delay)
        if rts_flip or dtr_flip:
            if dtr_rts >= 0:
                ser.dtr = dtr_flip ^ dtr
                await anyio.sleep(dtr_rts)
                ser.rts = rts_flip ^ rts
            else:
                ser.rts = rts_flip ^ rts
                await anyio.sleep(-dtr_rts)
                ser.dtr = dtr_flip ^ dtr
            await anyio.sleep(delay_flip)
        if dtr_rts > 0:
            ser.dtr = dtr
            await anyio.sleep(dtr_rts)
            ser.rts = rts
        else:
            ser.rts = rts
            await anyio.sleep(-dtr_rts)
            ser.dtr = dtr

        # flush messages
        if (t := cfg.get("flush")) is not None:
            if t is True:
                t = 0.2
            else:
                t /= 1000
            while True:
                with anyio.move_on_after(t):
                    res = await ser.receive(200)
                    log("Flush: %r", res)
                    continue
                break
        return ser



"""
Characer-based ring buffer.
"""

from __future__ import annotations

from typing import cast


class RingBuffer:
    """
    This ring buffer can hold a predetermined number of bytes.

    Old data are overwritten when the reader doesn't keep up.
    """

    def __init__(self, length: int):
        """
        Args:
            length: Buffer length, *not* constrained to powers of two.
        """
        self._buf = bytearray(length)
        self._read_pos = 0  # Position to read from
        self._count = 0  # Number of bytes available

    @property
    def n_free(self):
        "free space in buffer"
        return len(self._buf) - self._count

    @property
    def n_avail(self):
        "bytes in buffer"
        return self._count

    def __len__(self):
        return self._count

    def __repr__(self):
        return f"<Ring:{self._count}/{len(self._buf)}>"

    def write(self, buf: bytes, drop: bool = True) -> int:
        """
        Adds the bytes in `buf` to the end of the buffer.

        If the buffer is full and `drop` is set (the default), discards as
        many old bytes as necessary and replaces the first byte with a
        0x00. Otherwise only write as much as will fit.
        """
        if not buf:
            return 0

        buf_len = len(self._buf)
        res = write_len = len(buf)

        # If writing more than buffer can hold, only write last buf_len bytes
        if drop and write_len > buf_len:
            buf = cast(bytes, memoryview(buf)[write_len - buf_len :])
            write_len = buf_len

        # Check if we'll overflow
        overflow = max(0, self._count + write_len - buf_len)
        if overflow > 0:
            # Advance read position to discard old bytes
            if drop:
                self._read_pos = (self._read_pos + overflow) % buf_len
                self._count -= overflow
            else:
                res = write_len = buf_len - self._count
                overflow = 0
                buf = cast(bytes, memoryview(buf)[:write_len])

        # Write data, handling wraparound
        write_pos = (self._read_pos + self._count) % buf_len
        space_to_end = buf_len - write_pos
        if write_len <= space_to_end:
            # Simple case: no wraparound
            self._buf[write_pos : write_pos + write_len] = buf
        else:
            # Wraparound case: split the write
            self._buf[write_pos:buf_len] = buf[:space_to_end]
            self._buf[0 : write_len - space_to_end] = buf[space_to_end:]

        self._count += write_len

        # Replace the oldest byte with 0x00 if we had overflow
        if overflow > 0:
            self._buf[self._read_pos] = 0x00

        return res

    def readinto(self, buf: bytearray) -> int:
        """
        Copies as many bytes as will fit (or are available, whichever is
        smaller) into the buffer and advance the read counter.
        """
        read_len = min(len(buf), self._count)
        if read_len == 0:
            return 0

        buf_len = len(self._buf)
        space_to_end = buf_len - self._read_pos

        if read_len <= space_to_end:
            # Simple case: no wraparound
            buf[:read_len] = self._buf[self._read_pos : self._read_pos + read_len]
        else:
            # Wraparound case: split the read
            buf[:space_to_end] = self._buf[self._read_pos : buf_len]
            buf[space_to_end:read_len] = self._buf[0 : read_len - space_to_end]

        self._read_pos = (self._read_pos + read_len) % buf_len
        self._count -= read_len
        return read_len



"""
Steinhart-Hart NTC thermistor formulae
"""

from __future__ import annotations

from math import exp, log

NOMINAL_TEMPERATURE = 25
ADC_BITS = 10

__ALL__ = ["thermistor2celsius", "celsius2thermistor"]


def thermistor2celsius(B: int, raw: int, bits: int = ADC_BITS) -> float | None:
    """
    Convert thermistor measurement to temperature.

    source: https://arduinodiy.wordpress.com/2015/11/10/
        /measuring-temperature-with-ntc-the-steinhart-hart-formula/
    """
    if raw == 0:
        return None

    res = ((1 << bits) - 1) / raw - 1.0
    res = log(res) / B + 1 / (NOMINAL_TEMPERATURE + 273.15)
    res = 1.0 / res - 273.15

    return res


def celsius2thermistor(B: int, degC: float | None, bits: int = ADC_BITS) -> int:
    """
    Convert temperature to thermistor measurement.
    """
    if degC is None:
        return 0
    res = 1 / (degC + 273.15) - 1 / (NOMINAL_TEMPERATURE + 273.15)
    res = exp(res * B)
    res = ((1 << bits) - 1) / (res + 1.0)

    return int(res + 0.5)



"""
R/W access to configuration data.
"""

from __future__ import annotations

from moat.util import ExpKeyError, NotGiven, attrdict, enc_part, get_part, to_attrdict
from moat.lib.micro import log as log
from moat.lib.rpc import BaseCmd


class Cmd(BaseCmd):
    """
    Subsystem to handle config data.

    This app serves the config of the parent subcommand.
    """

    doc = dict(_c=dict(_d="Config read/write"))

    def __init__(self, cfg):
        super().__init__(cfg)
        self.repeats = {}

    doc_r = dict(_d="read cfg", _r="parts:dict[str,Any]:Config data")

    async def stream_r(self, msg):
        """
        Read (part of) the configuration.

        As configuration data are frequently too large to transmit in one
        go, this code interrogates them step-by-step.

        Args:
            p: path. An empty path is the root.

        If the accessed item is a dict, return data consists of a dict
        (simple keys and their values) and a list (keys for complex
        values).

        Likewise, the data for a list is a list with "simple" elements
        (complex data replaced by `None`), plus a list with the offsets of
        complex items that should be retrieved recursively.
        """
        p = msg.get("p", ())
        try:
            res = enc_part(get_part(self.root.cfg, p))
            if isinstance(res, (list, tuple)):
                await msg.result(*res)
            else:
                await msg.result(res)
        except KeyError as exc:
            raise ExpKeyError(*exc.args) from None

    doc_c = dict(_d="write cfg", p="Path:subpart", d="any:Data")

    async def cmd_c(self, p, d=NotGiven, keep: bool = False):
        """
        Online configuration mangling.

        As configuration data are frequently too large to transmit in one
        go, this code updates them step-by-step.

        Args:
            p: data path. It cannot be empty. Destinations are
               autogenerated. A path element of `None`, if last, appends
               to a list.

            d: data replacing the destination. `NotGiven` (or omitting the
               parameter) deletes.

        There is no way to write the current config to the file system.
        You can assemble it on the server and write using app.fs, or you
        can configure a "safe" skeleton setup and update it online after
        booting.
        """
        cur = self.root.cfg
        if not p:
            raise ValueError("NoPath")
        for pp in p[:-1]:
            try:
                cur = cur[pp]
            except KeyError:
                cur[pp] = attrdict()
                cur = cur[pp]
            except (IndexError, TypeError):
                if pp is not None and len(cur) != pp:
                    raise
                cur.append(attrdict())
                cur = cur[-1]
        # log("CFG_W %r %r %r", cur, p, d)
        k = p[-1]
        d = to_attrdict(d)
        if d is NotGiven and not keep:
            try:
                del cur[k]
            except KeyError:
                pass
        elif isinstance(cur, list) and k is None:
            cur.append(d)
        else:
            try:
                cur[k] = d
            except IndexError:
                if len(cur) != k:
                    raise
                cur.append(d)

        self.root.cfg_updated(cur)

    doc_x = dict(_d="activate new config")

    async def cmd_x(self):
        """
        Activate the new config.
        """
        await self.root.reload()



"""
Watchdog timer.

Config options:

* timeout

  Minimum time between "ping"s. Zero turns off the watchdog. Depending on
  the hardware you may or may not be able to change this, once it's >0.

* ext

  Require external keepalives.

* hw

  Use the singleton hardware watchdog.

"""

from __future__ import annotations

import sys

import machine

import moat.micro.main as M
from moat.lib.micro import Event, TimeoutError, wait_for_ms  # noqa:A004

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from typing import Never

try:
    _reset = machine.reset
except AttributeError:  # on Linux?
    import ffi

    C = ffi.open(None)

    def _reset():
        C.func("i", "_exit", "i")(11)


class WDT:
    """
    A wrapper for hardware/software watchdog Timer(s).

    Config:

        t: timeout in msec
        hw: flag whether to use the hardware watchdog
        ext: flag whether to require external ping from the server
        tt: trigger timeout, defaults to t/2
    """

    wdt = None
    timeout = None
    _ping = None

    def __init__(self, cfg):
        self.cfg = cfg
        if cfg.get("hw", False):
            if M.WDT is not None:
                raise RuntimeError("one hw watchdog")
            M.WDT = self
        self._ping = Event()

    def setup(self):  # noqa:D102
        cfg = self.cfg
        t = cfg.get("t", 0)
        if not t:
            if self.wdt is None:
                self.timeout = 0
        elif self.wdt:
            pass  # can't change or turn off

        else:
            if self.cfg.get("hw", False):
                self.wdt = machine.WDT(0, t)
            self.timeout = t
            self.trigger = self.cfg.get("tt", self.timeout / 2)
        self._ping.set()

    async def run(self) -> Never:  # noqa:D102
        T = getattr(machine, "Timer", None)
        t = None
        while True:
            if self.cfg.get("ext", False):
                if self.wdt is not None:
                    # wait for external trigger
                    # the HW WDT will kill us if it doesn't arrive
                    await self._ping.wait()
                elif self.timeout > 0:
                    # die hard if the external ping doesn't arrive
                    try:
                        await wait_for_ms(self.timeout, self._ping.wait)
                    except TimeoutError:
                        print("\n\nPANIC WDT REBOOT\n", file=sys.stderr)
                        for _ in range(10000):
                            pass
                        _reset()
                else:
                    await self._ping.wait()
            elif self.wdt is not None:
                # feed the watchdog when the trigger expires
                try:  # no "with suppress" on µPy
                    await wait_for_ms(self.trigger, self._ping.wait)
                except TimeoutError:
                    pass
            else:
                # use a software timeout
                if t is None and T is not None:
                    t = T(self.cfg.get("timer", -1))
                if t is None:
                    raise RuntimeError("no timer")
                if self.timeout:
                    t.init(
                        period=self.timeout,
                        mode=T.ONE_SHOT,
                        callback=lambda _: _reset(),
                    )
                try:
                    await self._ping.wait()
                finally:
                    t.deinit()

            if self.wdt is not None:
                self.wdt.feed()
            if self._ping.is_set():
                self._ping = Event()

    def tmo(self, timeout=None):
        """
        get/set the timeout
        """
        t = self.timeout
        if timeout is not None:
            self.timeout = timeout
            self._ping.set()
        return t

    def ping(self, force=False):
        """
        Keepalive trigger for the watchdog.

        If @force is set, the timer is set directly; if not, the watchdog
        task is woken up.
        """
        if force and self.wdt is not None:
            self.wdt.feed()
        if self._ping is not None:
            self._ping.set()



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

from moat.lib.micro import (
    TimeoutError,  # pylint: disable=redefined-builtin # noqa:A004
    log,
    wait_for_ms,
)

# Typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

__all__ = ["de_async", "wait_complain"]


class OutOfData(EOFError):
    pass


async def wait_complain(s: str, i: int, p: Callable, *a, **k):
    """
    Wait for a callable to complete, complaining if it takes too long.

    Args:
        s: Description string for logging.
        i: Timeout in milliseconds.
        p: Callable to wait for.
        *a: Positional arguments for the callable.
        **k: Keyword arguments for the callable.
    """
    try:
        await wait_for_ms(i, p, *a, **k)
    except TimeoutError:
        log("Delayed  %s", s)
        await p(*a, **k)
        log("Delay OK %s", s)


def de_async(p, *a, **k) -> Any:
    """
    Call an async proc that doesn't sleep.
    """
    try:
        p(*a, **k).send(None)
    except StopIteration as err:
        return err.value
    raise RuntimeError(f"{p} did sleep!")



from __future__ import annotations

from moat.micro.cmd import BaseCmd


class BMSCmd(BaseCmd):
    def __init__(self, parent, name, cfg):
        super().__init__(parent, name)
        self.cfg = cfg

    async def run(self):
        try:
            await self.batt.run(self)
        finally:
            self.batt = None

    async def config_updated(self, cfg):
        await super().config_updated(cfg)
        await self.batt.config_updated(cfg)

    doc_rly = dict(
        _d="relay get/force-set", st="bool?:forced state", _r=["bool:state", "bool:forced?"]
    )

    async def cmd_rly(self, msg):
        """
        Force the relay (st=bool), un-force it (st=None), or return the
        current state (st missing).

        Called manually, but also irreversibly when there's a "hard" cell over/undervoltage
        """
        if self.batt.relay is None:
            raise RuntimeError("No Relay")
        if "st" in msg:
            await self.batt.set_relay_force(msg["st"])
        else:
            return self.batt.relay.value(), self.batt.relay_force

    loc_rly = cmd_rly

    doc_info = dict(_d="incr status", gen="int:old state", r="bool:reset", _r="dict")

    async def cmd_info(self, msg):
        if self.bms.gen == msg.get("gen", -1):
            await self.bms.xmit_evt.wait()
        return self.bms.stat(msg.get("r", False))

    loc_info = cmd_info

    def cmd_live(self):
        self.bms.set_live()

    loc_live = cmd_live



"""
TCP single-connection listener app.
"""

from __future__ import annotations

from moat.lib.rpc import BaseListenOneCmd
from moat.lib.rpc.conn.tcp import TcpIter


class LinkIn(BaseListenOneCmd):
    """
    An app that accepts a single connection from a remote socket.

    New connections may or may not supersede existing ones, depending on the
    "replace" config item.
    """

    def listener(self):
        """Returns the TCP listener."""
        return TcpIter(self.cfg.get("host", "127.0.0.1"), self.cfg["port"])



"""
Helpers for MoaT command interpreters et al.
"""

from __future__ import annotations

# Typing

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import Protocol

    class HasValueTaskReplies(Protocol):
        """Required reply methods on command objects."""

        async def reply_error(self, i: int, err: Exception, x: list[Exception]) -> None:
            """Forward a task error."""

        async def reply_result(self, i: int, res: object) -> None:
            """Forward a task result."""

    class Cancellable(Protocol):
        """Cancel-capable task handle."""

        def cancel(self) -> None:
            """Cancel this task."""


class ValueTask:
    """
    An object that forwards a task's return value.

    @i: seqnum
    @x: excluded errors
    @p: callable
    """

    def __init__(
        self,
        cmd: HasValueTaskReplies,
        i: int,
        x: list[Exception],
        p: Callable[..., Awaitable[object]],
        *a: object,
        **k: object,
    ):
        self.cmd = cmd
        self.i = i
        self.p = p
        self.a = a
        self.k = k
        self.x = x
        self._t: Cancellable | None = None

    async def start(self, tg):
        "Task starter. Called from the command."
        if self._t is not None:
            raise RuntimeError("dup")
        self._t = await tg.spawn(self._wrap, _name="Val")

    async def _wrap(self):
        try:
            res = await self.p(*self.a, **self.k)
        except Exception as err:  # pylint:disable=broad-exception-caught
            await self.cmd.reply_error(self.i, err, self.x)
            return
        await self.reply_result(res)

    async def reply_result(self, res):
        "forward the task's return value to the caller"
        await self.cmd.reply_result(self.i, res)

    def cancel(self):
        "cancel the iterator"
        if self._t is not None:
            self._t.cancel()
            self._t = None

    async def set_error(self, err):
        "tell the iterator to raise an error"
        self.cancel()
        await self.cmd.reply_error(self.i, err, self.x)



"""
Test implementation of something that may or may not behave like a battery
"""

from __future__ import annotations

import logging
from functools import partial

from moat.util import attrdict
from moat.lib.micro import Queue, TaskGroup
from moat.lib.rpc import BaseCmd
from moat.micro.app.bms._test.diy_packet import PacketHeader, replyClass

logger = logging.getLogger(__name__)


class _CellSim(BaseCmd):
    # needs "ctrl" and "cell" attributes
    ctrl = None
    cell = None

    async def task(self):
        while True:
            msg = await self.ctrl.xrb()
            hdr, msg = PacketHeader.decode(msg)
            addr = hdr.hops
            hdr.hops += 1
            if hdr.start > addr or hdr.start + hdr.cells < addr:
                # not for us
                await self.ctrl.xsb(m=hdr.encode_one(msg))
                continue

            hdr.seen = True

            pkt, msg = hdr.decode_one(msg)
            logger.debug("MSG %r %r", hdr, pkt)
            if pkt is not None:
                await pkt.to_cell(self.cell)
            rsp = replyClass[hdr.command]()
            await rsp.from_cell(self.cell)
            await self.ctrl.xsb(m=hdr.encode_one(msg, rsp))


class CellSim(_CellSim):
    """
    Back-end to simulate a single cell.

    This is a background app. It reads byte blocks from the loopback app at @ctrl,
    analyzes them, and replies according to the cell app at @cell.
    """

    async def setup(self):
        await super().setup()
        self.cell = self.root.sub_at(self.cfg["cell"])
        self.ctrl = self.root.sub_at(self.cfg["ctrl"])

    async def task(self):
        self.set_ready()
        await super().task()


class _SingleCellSim(_CellSim):
    """
    Interface for a cell in a series, configured via CellsSim.
    """

    def __init__(self, cell, ctrl):
        self.cell = cell
        self.ctrl = ctrl


class CellsSim(_CellSim):
    """
    Back-end to simulate multiple cells.

    Config:
        n: number of cells
        ctrl: LoopLink taking to them
        cell: path to the array of Cell objects this app shall control
    """

    def __init__(self, cfg):
        super().__init__(cfg)
        self.n_cells = cfg["n"]

    async def setup(self):
        await super().setup()
        self.ctrl = self.root.sub_at(self.cfg["ctrl"])

    async def task(self):
        cell = self.cfg["cell"]

        def _mput(q, m):
            return q(m)

        async with TaskGroup() as tg:
            q = None
            for i in range(self.n_cells):
                c = attrdict()
                if i == 0:  # first
                    c.xrb = self.ctrl.xrb
                else:
                    c.xrb = q.get
                if i < self.n_cells - 1:
                    q = Queue()
                    c.xsb = partial(_mput, q.put)
                else:  # last
                    c.xsb = self.ctrl.xsb

                cp = self.root.sub_at(cell / i)
                sim = _SingleCellSim(cp, c)
                await tg.spawn(sim.task)
            self.set_ready()



"""
Random parts
"""

from __future__ import annotations

_attrs = {
    "PID": "pid",
    "Pin": "pin",
    "PWM": "pwm",
    "NoOp": "noop",
    "Relay": "relay",
    "Transfer": "transfer",
    "Average": "average",
}


# Lazy loader, effectively does:
#   global attr
#   from .mod import attr
def __getattr__(attr):
    mod = _attrs.get(attr, None)
    if mod is None:
        raise AttributeError(attr)
    value = getattr(__import__(f"moat.micro.part.{mod}", globals(), None, (attr,), 0), attr)
    globals()[attr] = value
    return value



"""
Stream infrastructure for handling data streams in a structured manner.
"""

from __future__ import annotations

# Import base classes directly (not lazy)
from .base import Base as Base
from .base import BaseBlk as BaseBlk
from .base import BaseBuf as BaseBuf
from .base import BaseConn as BaseConn
from .base import BaseMsg as BaseMsg
from .base import StackedBlk as StackedBlk
from .base import StackedBuf as StackedBuf
from .base import StackedConn as StackedConn
from .base import StackedMsg as StackedMsg

from typing import TYPE_CHECKING as _TC

if _TC:
    from .asyncio import AIOBuf as AIOBuf
    from .asyncio import SingleAIOBuf as SingleAIOBuf
    from .cbor import CBORMsgBlk as CBORMsgBlk
    from .cbor import CBORMsgBuf as CBORMsgBuf
    from .log import LogBlk as LogBlk
    from .log import LogBuf as LogBuf
    from .log import LogMsg as LogMsg
    from .reliable import EphemeralMsg as EphemeralMsg
    from .reliable import ReliableMsg as ReliableMsg
    from .serial import serial_stack as serial_stack
    from .tcp import TcpLink as TcpLink


# Compatibility
class ProcessDeadError(Exception):
    "unused, compatibility"

    pass


# Lazy loading
_imports = {
    # Serial stack
    "serial_stack": "serial",
    # Logging
    "LogMsg": "log",
    "LogBlk": "log",
    "LogBuf": "log",
    # CBOR
    "CBORMsgBuf": "cbor",
    "CBORMsgBlk": "cbor",
    # Asyncio
    "AIOBuf": "asyncio",
    "SingleAIOBuf": "asyncio",
    # Reliable messaging
    "ReliableMsg": "reliable",
    "EphemeralMsg": "reliable",
    # Network connections
    "TcpLink": "tcp",
}


def __getattr__(attr: str):
    try:
        mod = _imports[attr]
    except KeyError:
        raise AttributeError(attr) from None
    value = getattr(__import__(mod, globals(), None, True, 1), attr)
    globals()[attr] = value
    return value


__all__ = [  # noqa:RUF022
    # Base classes (not lazy)
    "Base",
    "BaseBlk",
    "BaseBuf",
    "BaseConn",
    "BaseMsg",
    "StackedBlk",
    "StackedBuf",
    "StackedConn",
    "StackedMsg",
    # Logging (lazy)
    "LogBlk",
    "LogBuf",
    "LogMsg",
    # CBOR (lazy)
    "CBORMsgBlk",
    "CBORMsgBuf",
    # Asyncio (lazy)
    "AIOBuf",
    "SingleAIOBuf",
    # Reliable messaging (lazy)
    "ReliableMsg",
    "EphemeralMsg",
    # Serial
    "serial_stack",
    # Network connections (lazy)
    "TcpLink",
    # compatibility
    "ProcessDeadError",
]



"""
Support for various ICs.

This module uses lazy loading to avoid circular dependencies.
"""

from __future__ import annotations

from typing import TYPE_CHECKING as _TC

if _TC:
    from .hdc2010 import Cmd as HDC2010
    from .tmp102 import Cmd as TMP102

_imports = {
    "HDC2010": ("hdc2010", "Cmd"),
    "TMP102": ("tmp102", "Cmd"),
}


def __getattr__(attr: str):
    try:
        mod, name = _imports[attr]
    except KeyError:
        raise AttributeError(attr) from None
    value = getattr(__import__(mod, globals(), None, (name,), 1), name)
    globals()[attr] = value
    return value


def __dir__():
    """Expose all lazy-loaded attributes for introspection."""
    return sorted(set(list(__all__) + [k for k in globals().keys() if not k.startswith("_")]))


__all__ = ["HDC2010", "TMP102"]



"""
When a channel is lossy, this module implements re-sending messages.
"""

from __future__ import annotations

from contextlib import suppress

from moat.util import NotGiven, Queue, ValueEvent
from moat.lib.codec.errors import ChannelClosed
from moat.lib.micro import (
    ACM,
    AC_exit,
    Event,
    TaskGroup,
    TimeoutError,  # noqa:A004
    idle,
    log,
    sleep_ms,
    ticks_add,
    ticks_diff,
    ticks_ms,
    wait_for_ms,
)
from moat.lib.stream import StackedMsg

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any  # isort:skip


class EphemeralMsg:
    """A message that may be replaced while enqueued or in transit.

    Sending an `EphemeralMsg` enqueues the message without waiting for
    delivery. If a channel's message is waiting to be sent, it'll be
    updated.
    """

    def __init__(self, chan: int, data: Any):
        self.chan = chan
        self.data = data


class ReliableMsg(StackedMsg):
    """
    Message retry.

    This module handles retransmitting missing messages.
    """

    # Operation:
    #
    # Both sender and receiver carry a window-sized array of messages and
    # two position markers: head and tail.
    #
    # Sending, "head" is the next message to be sent, "tail" is the
    # oldest message that's not yet acknowledged. "tail" is populated (if
    # different from "head"), "head" is not, those in between may not be if
    # they have been selectively acknowledged.
    #
    # Receiving, "head" is the next message we expect, "tail" is the oldest
    # message we have not received yet. The data at both "head" and "tail"
    # must be empty (obviously). Between those there may be messages that
    # have been received out of order.
    #
    # Both sender and receiver limit the difference from tail to head to
    # window/2. The sender delays messages that would exceed the window.
    # The receiver discards messages outside of this range.
    #
    # All messages are lists and contain the send_position `s`, recv_tail
    # `r`, recv_done `x`, and data `d`. "recv_done" is a list of messages
    # that have been received out-of-order so they're not retransmitted.
    #
    # 'x' is only present if 'r' is negated (``-1-r``).
    #
    # The receiver advances its send_tail until it matches `r` and queues
    # all messages for retransmission that are still outstanding but not
    # mentioned in `x`. `x` is omitted when empty, `d` is omitted when not
    # yet queued.
    #
    # If data is a list, they're appended as-is to the message, to save a
    # byte, except when the list is one element long, for data
    # transparency.
    #
    # Sending an EphemeralMsg does not wait for delivery. Instead, the
    # message is queued. If the message has not been sent by the time the
    # channel's next message is transmitted, it is updated instead.
    #
    # Connection reset or restart is signalled by `s`<0. Values -2 to -4
    # correspond to restart phases; the following elements contain link
    # configuration data. A value of `-1` indicated that the connection is
    # closed. Other values are reserved.
    #
    # `c` contains of at least two values: the window size and a timeout,
    # in milliseconds. When values do not match, max timeout and min
    # window values are used by both sides. Other items are ignored.
    # A value of zero (or a missing parameter) means "OK I'll take your
    # value" and may only be sent after the corresponding parameter has
    # been seen from the remote.
    #
    # The minimum window size is 4. The minimum timeout is 10 msec, which
    # is probably not useful. The default is 1000 (one second).

    rq = None
    _tg = None
    __tg = None

    def __init__(self, link, cfg):
        super().__init__(link, cfg)

        window = cfg.get("window", 8)
        timeout = cfg.get("timeout", 1000)
        retries = cfg.get("retries", 5)

        if window < 4:
            raise RuntimeError(f"window must be >=4, not {window}")
        self.window = window
        self.in_reset = None
        self.timeout = timeout
        self._trigger = Event()
        self._is_up = Event()
        self._is_down = Event()
        self._is_down.set()
        self.retries = retries
        self._iters = {}

    def reset(self, level=1):  # noqa:D102
        self.s_send_head = 0  # next to be transmitted
        self.s_send_tail = 0  # no open messages before this point
        self.s_recv_head = 0  # next expected message. Messages before this are out of sequence
        self.s_recv_tail = 0  # messages before this have been processed
        self.s_q = []
        self.m_send: dict[int, list[Any]] = {}
        # mte: message timestamp event
        self.m_recv = {}
        self.t_recv = None
        self.progressed = False
        self.in_reset = ticks_ms()
        self.reset_level = level
        self.pend_ack = True
        if self.rq is None:
            self.rq = Queue(self.window)

    async def send_msg(self, k=None):
        """
        Send some message.

        @k: index of whatever to send next, if unacknowledged;
        `None` if only sending an ACK.
        """
        self.progressed = True
        d = NotGiven
        if k is None:
            if not self.pend_ack:
                return
            msg = [self.s_send_head]
        else:
            mte = self.m_send[k]
            if mte[1] is False:
                # already ack'd.
                msg = [self.s_send_head]
            else:
                mte[1] = ticks_add(ticks_ms(), self.timeout)
                msg = [k]
                d = mte[0]
        r = self.s_recv_tail
        x = []
        while r != self.s_recv_head:
            if r in self.m_recv:
                x.append(r)
            r = (r + 1) % self.window
        if x:
            msg.append(-1 - self.s_recv_tail)
            msg.append(x)
        else:
            msg.append(self.s_recv_tail)
        if d is not NotGiven:
            if isinstance(d, EphemeralMsg):
                d.sent = True
                d = d.data
            if isinstance(d, (tuple, list)) and len(d) != 1:
                msg.extend(d)
            else:
                msg.append(d)
        self.pend_ack = False

        with suppress(RuntimeError):
            await self.s.send(msg)

        if k is not None and self.m_send.get(k, None) is mte:
            mte[1] = ticks_add(ticks_ms(), self.timeout)

    async def _run_bg(self):
        while not self.closed:
            if self.in_reset:
                await self._trigger.wait()
                self._trigger = Event()
                continue
            t = ticks_ms()
            # calculate time to next action
            ntx = None if self.t_recv is None else ticks_diff(self.t_recv, t)
            nk = None
            for k, mte in self.m_send.items():
                _m, tx, _e = mte
                if tx is None or tx is False:
                    continue
                txd = ticks_diff(tx, t)
                if ntx is None or ntx > txd:
                    ntx = txd
                    nk = k

            w_open = (
                self.s_q and (self.s_send_head - self.s_send_tail) % self.window < self.window // 2
            )
            if w_open:
                # yes, we can send another message
                # print(f"R {self.link.txt}: tx", file=sys.stderr)
                pass
            elif ntx is None:
                # nothing to do
                # print(f"R {self.link.txt}: inf", file=sys.stderr)
                await self._trigger.wait()
                self._trigger = Event()
            elif ntx > 0:
                # we need to do something soon
                # print(f"R {self.link.txt}: {ntx}", file=sys.stderr)
                try:
                    await wait_for_ms(ntx, self._trigger.wait)
                except TimeoutError:
                    pass
                else:
                    self._trigger = Event()

                # recalculate, as it may have changed during the wait
                w_open = (
                    self.s_q
                    and (self.s_send_head - self.s_send_tail) % self.window < self.window // 2
                )
                # XXX this prevents the "clash" error below but is probably
                # not the real fix
            else:
                # we need to re-send now
                # print(f"R {self.link.txt}: now {ticks_ms()}", file=sys.stderr)
                pass

            # process pending-send queue
            if w_open:
                seq = self.s_send_head
                msg, evt = self.s_q.pop(0)
                nseq = (seq + 1) % self.window
                # print("SH1",self.link.txt,self.s_send_tail,
                #             self.s_send_head,nseq, file=sys.stderr)
                self.s_send_head = nseq
                if seq in self.m_send:
                    raise RuntimeError("Clash")
                self.m_send[seq] = [msg, None, evt]
                await self.send_msg(seq)

            if ntx is not None and ntx <= 0:
                # send a retransmission or a pending ack
                if nk is not None:  # retransmit message K
                    await self.send_msg(nk)
                elif self.pend_ack:
                    await self.send_msg()

                if nk is None:
                    self.t_recv = ticks_add(ticks_ms(), self.timeout)

    async def wait(self):
        """
        Wait until the lower side is (again) ready to be used.

        There is no guarantee that this is still so when you call
        send/recv next.
        """
        while self.closed:
            await self._is_up.wait()

    async def __aenter__(self):
        acm = ACM(self)

        try:
            tg = await acm(TaskGroup())
            if self.retries:
                await tg.spawn(self._mon)
            await tg.spawn(self._run)
            if not self.cfg.get("_nowait"):  # required for testing
                await self.wait()
            self._tg = tg
            return self
        except BaseException as exc:
            await AC_exit(self, type(exc), exc, None)
            raise

    async def __aexit__(self, *err):
        if self._tg is not None:
            self._tg.cancel()
        self._tg = None
        return await AC_exit(self, *err)

    async def _mon(self):
        while True:
            await wait_for_ms(self.retries * self.timeout, self._is_up.wait)
            await self._is_down.wait()

    async def _run_(self):
        if self._is_down.is_set():
            self._is_down = Event()
        self.reset_level = 1

        try:
            link = self.link
            if link is None:
                raise RuntimeError("No link")
            async with TaskGroup() as tg, link as s:
                self.__tg = tg
                self.s = s
                await tg.spawn(self._read, _name="rel_read")
                await tg.spawn(self._run_bg, _name="rel_bg")
                while self.in_reset:
                    t = ticks_ms()
                    td = ticks_diff(self.in_reset, t)
                    if td > 0:
                        with suppress(TimeoutError):
                            await wait_for_ms(td, self._trigger.wait)
                        # DO NOT replace self._trigger here. _run_bg() already does that.
                    else:
                        await self.send_reset()

                if self.closed:
                    raise EOFError(self)

                await idle()

        except Exception:  # noqa:TRY203
            # if not self.persist:
            raise
            # log("Reliable", err=exc)

    async def _run(self):
        self.reset()
        try:
            while True:
                await self._run_()
                await sleep_ms(self.timeout)
        except BaseException as exc:
            err = str(exc)
            raise
        else:
            err = None
        finally:
            self.__tg = None
            self._is_down.set()
            for _m, _t, e in self.m_send.values():
                if e is not None:
                    e.set_error(ChannelClosed())
            while self.s_q:
                _m, e = self.s_q.pop()
                if e is not None:
                    e.set_error(ChannelClosed())
            if self._is_up.is_set():
                self._is_up = Event()
            msg = [-1]
            if err is not None:
                msg.append(err)
            try:
                try:
                    await self.s.send(msg)
                except AttributeError:
                    return  # closing. XXX should not happen
                except TypeError:
                    if len(msg) > 1:
                        msg = [msg[0], repr(err)]
                        await self.s.send(msg)
                    else:
                        raise
            except EOFError:
                pass

    async def send_reset(self, level=None, err=None):
        "send a Reset message to the other side"
        if level is None:
            level = self.reset_level
        else:
            self.reset_level = level
        msg = [-1 - level]
        if level:
            msg.extend(self._get_config())
        elif err is not None:
            msg.append(err)
        if self.reset_level < 3:
            self.in_reset = ticks_add(ticks_ms(), self.timeout)
        self._trigger.set()
        await self.s.send(msg)

    async def send(self, m: Any) -> Any:
        """
        Sender.

        Ephemeral messages get cached and updated.
        The sender won't wait until they're transmitted.
        """

        if isinstance(m, EphemeralMsg):
            i = m.chan
            if (om := self._iters.get(i, None)) is not None:
                om.data = m.data
                om.sent = False
                return
            self._iters[i] = m
            m.sent = False
            evt = None
        else:
            evt = ValueEvent()

        self.s_q.append((m, evt))
        self._trigger.set()
        if evt is None:
            return
        try:
            return await evt.wait()
        except BaseException:
            # TODO send a cancellation
            # self.s_q.append(({"i": msg["i"]}, None))
            self._trigger.set()
            raise

    async def error(self, exc: BaseException) -> None:
        """Report an internal protocol error."""
        log("Reliable error: %r", exc)

    def _get_config(self):
        return [self.window, self.timeout]

    def _update_config(self, c):
        if len(c) > 0 and c[0] > 0:
            self.window = max(4, min(self.window, c[0]))
        if len(c) > 1 and c[1] > 0:
            self.timeout = max(10, self.timeout, c[1])

    def _reset_done(self):
        if self.in_reset:
            self.in_reset = False
            self._trigger.set()
            self._is_up.set()

    async def recv(self) -> Any:
        "return the next message in the receive queue"
        if self.rq is None:
            raise EOFError
        return await self.rq.get()

    async def _read(self) -> None:
        while True:
            try:
                if self.s is None:
                    return
                msg = await self.s.recv()
            except EOFError:
                if self.__tg is not None:
                    self.__tg.cancel()
                return
            await self._dispatch(msg)

    @property
    def closed(self) -> bool:
        "check if the link is down"
        return self._is_down.is_set()

    async def _dispatch(self, msg):
        if not isinstance(msg, (tuple, list)):
            raise TypeError(msg)

        if msg[0] < 0:  # protocol error / reset sequence
            n = -1 - msg[0]
            if n == 0:  # closed
                self._is_down.set()
                if self._is_up.is_set():
                    self._is_up = Event()
                self._trigger.set()
            elif self.closed:
                await self.send_reset()
                return
            elif n == 1:  # incoming reset
                if self.in_reset:
                    if self.reset_level == 1:
                        self.reset_level = 2
                else:
                    self.reset(2)

                    await self.error(RuntimeError("ext reset"))
                self._update_config(msg[1:])
                await self.send_reset()
                return
            elif n == 2:  # incoming ack
                self._update_config(msg[1:])
                await self.send_reset(3)
                self._reset_done()
                return
            elif n == 3:  # incoming ack2
                if not self.in_reset or self.reset_level > 1:
                    self._update_config(msg[1:])
                    self._reset_done()
                else:
                    await self.error(RuntimeError("ext reset ack2"))
                    self.reset(1)
                    await self.send_reset()
                return
            else:
                log("Unknown", msg)
                return

        if self.in_reset:
            if self.reset_level < 2:
                await self.send_reset()
                return
            # assume that the other side got our reset=2
            # but their ack got lost
            self._reset_done()

        r = msg[0]  # swapped, because receiving
        s = msg[1]
        if s < 0:
            s = -1 - s
            x = msg[2]
            d = msg[3:]
        else:
            x = ()
            d = msg[2:]

        if not (0 <= r < self.window) or not (0 <= s < self.window):
            self.reset(1)
            await self.send_reset(err="R/S out of bounds")
            return

        if d:
            # data. R is the message's sequence number.
            self.pend_ack = True
            if self.between(self.s_recv_tail, self.s_recv_head, r):
                if (r - self.s_recv_tail) % self.window < self.window // 2:
                    # print("RH1",self.link.txt,self.s_recv_tail,
                    #             self.s_recv_head,r,r+1, file=sys.stderr)
                    self.m_recv[r] = d
                    self.s_recv_head = (r + 1) % self.window
                else:
                    pass
                    # print("RH1-",self.link.txt,self.s_recv_tail,
                    #             self.s_recv_head,r,r+1, file=sys.stderr)
            elif self.between(self.s_recv_tail, r, self.s_recv_head):
                self.m_recv[r] = d

        elif self.between(self.s_recv_tail, self.s_recv_head, r):
            # no data. R is the next-expected sequence number.
            if (r - self.s_recv_tail) % self.window <= self.window // 2:
                self.s_recv_head = r
                # print("RH2",self.link.txt,self.s_recv_tail,
                #             self.s_recv_head,r,r+1, file=sys.stderr)
            else:
                pass
                # print("RH2-",self.link.txt,self.s_recv_tail,
                #             self.s_recv_head,r,r+1, file=sys.stderr)

        # process ACKs
        if True:  # s >= 0:
            rr = self.s_send_tail
            while rr != s:
                if rr == self.s_send_head:
                    # XXX
                    break
                # log("ACKING %d",rr)
                try:
                    m, _t, e = self.m_send.pop(rr)
                except KeyError:
                    pass
                else:
                    if isinstance(m, EphemeralMsg):
                        mo = self._iters.pop(m.chan, None)
                        if not mo.sent:
                            # re-enqueue
                            self.s_q.append((msg, None))

                    self.pend_ack = True
                    if e is not None:
                        e.set(None)
                rr = (rr + 1) % self.window
                self._trigger.set()

            # print("ST1",self.link.txt,self.s_send_tail,self.s_send_head,rr, file=sys.stderr)
            self.s_send_tail = rr

        for rr in x:
            try:
                _m, _t, e = self.m_send[rr]
            except KeyError:
                pass
            else:
                self.m_send[rr][1] = False
                if e is not None:
                    e.set(None)

        # Forward incoming messages if s_recv[recv_tail] has arrived
        rr = self.s_recv_tail
        while rr != self.s_recv_head:
            try:
                d = self.m_recv.pop(rr)
            except KeyError:
                # missing message. Do not proceed.
                break
            else:
                rr = (rr + 1) % self.window
                self.s_recv_tail = rr
                # print("RT1",self.link.txt,self.s_recv_tail,
                #             self.s_recv_head,r,r+1, file=sys.stderr)
                self.pend_ack = True
                if len(d) == 1:
                    d = d[0]
                if self.rq is None:
                    raise EOFError
                await self.rq.put(d)

        if self.s_recv_tail == self.s_recv_head:
            self.t_recv = None
        else:
            self.t_recv = ticks_add(ticks_ms(), self.timeout)
            self._trigger.set()

        if self.pend_ack:
            # TODO delay ACK somewhat
            with suppress(EOFError):
                await self.send_msg()

    def between(self, a, b, c):
        "check if a,b,c are consecutive, modulo the window size"
        d1 = (b - a) % self.window
        d2 = (c - a) % self.window
        return d1 <= d2



"""
Test implementation of something that may or may not behave like a battery
"""

from __future__ import annotations

import logging
from math import exp

from moat.util import pos2val, val2pos
from moat.ems.battery._base import BalBaseCell
from moat.lib.micro import sleep_ms

logger = logging.getLogger(__name__)


class Cell(BalBaseCell):
    """
    Mock battery cell.

    Its nominal charge is 0…1, capacity in amp-sec.

    The voltages are linear between u.ext, plus a power term when exceeding
    that.

    A configuration that replicates a LiFePo4 cell, *very* approximately::

        c: 0.5
        t: 25
        cap: 2000
        i:
          dis: -1  # balancer discharge current
          chg: 0  # TODO
        lim:
          t:
            abs:
              min: 0
              max: 45
            ext:
              min: 10
              max: 40
          c:
            min: 0.2
            max: 0.9
          p:  # exponent when 'ext' limit is exceeded
            min: 2
            max: 3
          u:
            abs:
              min: 2.5
              max: 3.65
            std:
              min: 2.6
              max: 3.55
            ext:
              min: 2.9
              max: 3.4
    """

    def __init__(self, cfg):
        super().__init__(cfg)
        self._c = cfg["c"]
        self._t = cfg["t"]
        self._tb = 25
        self.t_env = 25

    async def set_dis(self):
        "set discharger. No-op since we can use .vdis directly."
        pass

    doc_c = dict(_d="get/set charge", c="float:new charge", _r="float:current charge")

    async def cmd_c(self, c: float | None = None) -> float:
        "get/set the current charge"
        if c is not None:
            self._c = c
        return self._c

    doc_u = dict(_d="get voltage", _r="float:voltage")

    async def cmd_u(self, c: float | None = None) -> float:
        """
        return the current voltage.

        If @c is set, return the voltage the cell would have if its current charge was c.
        """
        li = self.cfg["lim"]
        liu = li["u"]
        if c is None:
            c = self._c
        up = val2pos(li["c"]["min"], c, li["c"]["max"])
        if up < 0:
            fx = val2pos(li["c"]["min"], c, 0) ** li["p"]["min"]
            u = pos2val(liu["ext"]["min"], fx, liu["abs"]["min"])
        elif up > 1:
            fx = val2pos(li["c"]["max"], c, 1) ** li["p"]["max"]
            u = pos2val(liu["ext"]["max"], fx, liu["abs"]["max"])
        else:
            u = pos2val(liu["ext"]["min"], up, liu["ext"]["max"])
        return u

    doc_te = dict(_d="get/set env temp", c="float:new temp", _r="float:current temp")

    async def cmd_te(self, t=None) -> float:
        "gets/sets the environment temperature"
        if t is None:
            return self.t_env
        self.t_env = t

    doc_t = dict(_d="get balance temp", _r="float:current temp")

    async def cmd_t(self):
        "returns the balancer resistor temperature"
        return self._t

    doc_tb = dict(_d="get battery temp", _r="float:current temp")

    async def cmd_tb(self):
        "returns the battery temperature"
        return self._tb

    doc_add_p = dict(_d="add energy", p="float:watts", t="float:seconds")

    async def cmd_add_p(self, p, t):
        "add energy to the battery: @p watts for @t msec"
        # watt seconds
        self._c += p * t / self.cfg["cap"] / 1000

        # time takes heat away, Charge+Discharge adds it
        self._t += (self.t_env - self._t) * (1 - exp(-t / 10000)) + abs(p) * t / 100000 * (
            1 if p > 0 else 0.5
        )

    async def task(self):
        self.set_ready()

        while True:
            await sleep_ms(100)
            if self.vdis:
                u = await self.cmd_u()
                if u > self.vdis:
                    await self.cmd_add_p(u * self.cfg["i"]["dis"], 100)

            if self.vchg:
                u = await self.cmd_u()
                if u < self.vchg:
                    await self.cmd_add_p(u * self.cfg["i"]["chg"], 100)

    doc_bal_set = dict(
        _d="set balancer",
        pwm="any:PWM data",
        b="bool:in balance",
        f="nool:bal forced",
        ot="bool:balancer overtemp",
        th="float:bal temp threshold",
    )

    async def cmd_bal_set(self, b=None, f=None, ot=None, pwm=None, th=None):
        if pwm is not None:
            self.balance_pwm = pwm
        if ot is not None:
            self.balance_over_temp = ot
        if f is not None:
            self.balance_forced = f
        if th is not None:
            self.balance_threshold = th
        if b is not None:
            self.in_balance = b


class DiyBMSCell(Cell):
    bc_i = 1000
    bc_e = 2000
    v_per_ADC = 0.001
    n_samples = 3
    v_calibration = 1.0
    v_offset = 0

    doc_b_coeff = dict(_d="get coef", _r=["float:i", "float:e"])

    async def cmd_b_coeff(self):
        return self.bc_i, self.bc_e

    doc_v = dict(_d="calc raw volt", _0="float:voltage", _r="int:value")

    async def cmd_v2raw(self, val):
        if val is None or self.n_samples is None:
            return 0
        return int((val - self.v_offset) / self.v_per_ADC * self.n_samples / self.v_calibration)

    doc_raw = dict(_d="calc real volt", _r="float:voltage", _0="int:value")

    async def cmd_raw2v(self, val):
        if val is None or self.cfg.u.samples is None or val == 0:
            return None
        return val * self.v_per_ADC / self.cfg.u.samples * self.v_calibration + self.cfg.u.offset

    doc_settings = dict(
        _d="get settings",
        _r=dict(
            vpa="float:vol per ADC tick", ns="int:n samples", vcal="float:volt caibration factor"
        ),
    )

    async def cmd_settings(self):
        return dict(vpa=self.v_per_ADC, ns=self.n_samples, vcal=self.v_calibration)



"""
SerialPacker protocol support for stream layers.
"""

from __future__ import annotations

from moat.lib.micro import Lock
from moat.lib.stream import BaseBuf, BaseMsg, StackedBlk

from ._console import _CReader

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from moat.util import attrdict
    from moat.lib.stream.base import Buffer


class SerialPackerBlkBuf(StackedBlk):
    """
    chunked bytestrings > SerialPacker-ized stream

    Use this (and a CBORHandler and a Reliable) if your AIO stream
    is unreliable (TTL serial).
    """

    cons = False

    def __init__(self, stream: BaseBuf, frame: dict, console: bool | int = False):
        StackedBlk.__init__(self, stream, None)

        SerialPacker = __import__("serialpacker", fromlist=("SerialPacker",)).SerialPacker

        self.p = SerialPacker(**frame)
        self.buf = bytearray(16)
        self.i = 0
        self.n = 0
        self.w_lock = Lock()
        if console:
            _CReader.__init__(cast(_CReader, self), console)

    async def crd(self, buf: bytearray) -> int:
        "console read"
        if not self.cons:
            raise EOFError
        return await _CReader.crd(cast(_CReader, self), buf)

    async def cwr(self, buf) -> None:
        "console write"
        if not self.cons:
            return
        await self.s.wr(buf)

    async def rcv(self):
        "block read"
        while True:
            while self.i < self.n:
                msg = self.p.feed(self.buf[self.i])
                self.i += 1
                if isinstance(msg, int):
                    if self.cons:
                        _CReader.cput(cast(_CReader, self), msg)
                elif msg is not None:
                    return msg

            n = await self.s.rd(self.buf)
            if not n:
                raise EOFError
            self.i = 0
            self.n = n

    async def snd(self, m: Buffer) -> None:
        "block write"
        h, msg, t = self.p.frame(m)
        async with self.w_lock:
            if not self.cons:
                await self.s.wr(h)
                await self.s.wr(msg)
                await self.s.wr(t)
            else:
                await self.s.wr(h + msg + t)

    recv = rcv
    send = snd


def serial_stack(stream, cfg: attrdict, cons: bool = False):
    # lossy=False, log=False, use_console=False, msg_prefix=None
    """
    Build a message stack on top of a MoaT bytestream.

    Configuration:
        link(dict):
            Link control; see below.
        log(dict):
            If present, log high-level messages.
        log(dict):
            If present, log messages
        log_raw(dict):
            If present, log the bytestream.

    Link control:
        cbor(bool):
            must be ``True``.
        lossy(bool):
            set if the stream is not 100% reliable.
        frame(int|dict):
            control protocol framing.
        console(bool):
            set if incoming non-framed data should be processed.

    If ``lossy`` is `True`, ``frame`` must be a dict.

    If ``frame`` is an integer, the protocol is assumed to be
    self-delimiting (e.g. CBOR). If it's a dict, the value is
    the configuration for a ``SerialPacker`` instance.

    There is no start-of-frame character escaping. Choose a value that cannot occur
    in an ASCII or possibly UTF-8 bytestream (≥ 0xF8).
    """

    if not hasattr(stream, "rd") or not hasattr(stream, "wr"):
        raise TypeError(f"need a BaseBuf not {stream}")

    link = cfg.get("link", {})
    cons = link.get("console", cons)
    frame = link.get("frame", None)
    lossy = link.get("lossy", None)
    log = cfg.get("log", None)
    log_raw = cfg.get("log_raw", None)
    log_rel = cfg.get("log_rel", None)

    if log_raw is not None:
        from moat.lib.stream import LogMsg  # noqa: PLC0415

        stream = LogMsg(stream, log_raw)

    if isinstance(frame, dict):
        from moat.lib.stream import CBORMsgBlk  # noqa: PLC0415

        stream = SerialPackerBlkBuf(stream, frame=frame, console=cons)
        stream = CBORMsgBlk(stream, cfg)
    else:
        from moat.lib.stream import CBORMsgBuf  # noqa: PLC0415

        stream = CBORMsgBuf(stream, dict(msg_prefix=frame, console=cons))

    assert isinstance(stream, BaseMsg)

    if lossy:
        if lossy is True:
            lossy = {}
        from moat.lib.stream import ReliableMsg  # noqa: PLC0415

        if log_rel is not None:
            from moat.lib.stream import LogMsg  # noqa: PLC0415

            stream = LogMsg(stream, log_rel)

        stream = ReliableMsg(stream, lossy)

    if log is not None:
        from moat.lib.stream import LogMsg  # noqa: PLC0415

        stream = LogMsg(stream, log)

    return stream



"""
Command-line code for moat.micro
"""

# pylint: disable=import-outside-toplevel
from __future__ import annotations

import anyio
import logging
import os
import sys
from functools import wraps
from pathlib import Path as FSPath
from shutil import which

import asyncclick as click

from moat.util import (
    NotGiven,
    SigCancel,
    attrdict,
    combine_dict,
    get_part,
    merge,
    to_attrdict,
    yload,
    yprint,
)
from moat.lib.codec import get_codec
from moat.lib.codec.errors import NoPathError, RemoteError
from moat.lib.micro import idle
from moat.lib.path import (
    P,
)
from moat.lib.rpc import Msg, RootCmd
from moat.lib.run import attr_args, load_subgroup, process_args
from moat.micro.util import TEST_MAGIC

logger = logging.getLogger(__name__)


skip_exc = {FileNotFoundError, FileExistsError, ConnectionRefusedError}


def catch_errors(fn):
    """
    Wrapper for commands so that some errors don't cause a stack trace.
    """

    @wraps(fn)
    async def wrapper(*a, **k):
        try:
            return await fn(*a, **k)
        except (NoPathError, ConnectionRefusedError) as e:
            raise click.ClickException(str(e))  # noqa:B904
        except Exception as e:
            if "bdb" in sys.modules:
                skip_exc.add(sys.modules["bdb"].BdbQuit)
            if type(e) in skip_exc and "MOAT_TB" not in os.environ:
                raise click.ClickException(repr(e))  # noqa:B904
            raise

    return wrapper


@load_subgroup(
    prefix="moat.micro",
    epilog="""
        The 'section' parameter says which part of the MoaT configuration
        (below "moat.micro") to use to connect to the remote system.
        The defaults are:

        \b
        Command  Section
        =======  =======
        run      run        server mode
        setup,   setup      used for initial non-MoaT access to the device
         install
        *        connect    used for all other commands

        The 'remote' parameter specifies the prefix that talks to
        the remote system. The default is 'r'.

        """,
)
@click.pass_context
@click.option("-S", "--section", type=P, help="Section to use")
@click.option("-R", "--remote", type=P, help="Path for talking to the satellite")
@click.option("-P", "--path", type=(str, P), multiple=True, help="named remote component")
async def cli(ctx, section, remote, path):
    """Run MicroPython satellites

    'moat micro' configures MoaT satellites and runs the link to them,
    as well as applications using it."""
    obj = ctx.obj
    cfg = obj.cfg.micro
    remote2 = None
    if "--help" in sys.argv:
        return  # HACK

    inv = ctx.invoked_subcommand
    if inv == "run":
        if remote is not None:
            raise click.UsageError("You don't use a remote with 'moat micro run'")
        if path:
            raise click.UsageError("You don't use paths with 'moat micro run'")
        if section is None:
            section = P("run")
    elif inv in ("setup", "install"):
        if remote is not None:
            remote2, remote = remote, None
        else:
            remote2 = P("s")
        if path:
            raise click.UsageError("You don't use paths with 'moat micro setup|install'")
        if section is None:
            section = P("setup")
    elif inv == "path":
        pass
    else:
        if section is None:
            section = P("connect")
    if section is not None:
        try:
            cfg = get_part(cfg, section)
        except KeyError:
            raise click.UsageError(f"The config section '{section}' doesn't exist.") from None
    if remote2 is not None:
        cfg.remote2 = remote2

    if remote is not None:
        cfg.remote = remote
    elif "remote" not in cfg:
        cfg.remote = P("r")
    pth = cfg.setdefault("path", {})
    for n, p in path:
        pth[n] = p
    for k, v in pth.items():
        if isinstance(v, str):
            v = P(v)  # noqa:PLW2901
        pth[k] = v

    obj.mcfg = to_attrdict(cfg)


@cli.command(name="setup", short_help="Copy MoaT to MicroPython")
@click.pass_context
@click.option("-i", "--install", is_flag=True, help="Install MicroPython")
@click.option("-r", "--run", is_flag=True, help="Run MoaT after updating")
@click.option("-F", "--rom", is_flag=True, help="Upload to ROM Flash")
@click.option("--run-section", type=P, help="Section with runtime config (default: 'run')")
@click.option("-N", "--reset", is_flag=True, help="Reboot after updating")
@click.option("-K", "--kill", is_flag=True, help="Reboot initially")
@click.option(
    "-S",
    "--source",
    type=click.Path(dir_okay=True, file_okay=True, path_type=anyio.Path),
    help="Files to sync",
)
@click.option("-D", "--dest", type=str, default="", help="Destination path")
@click.option("-R", "--root", type=str, default=".", help="Destination root")
@click.option("-U/-V", "--update/--no-update", is_flag=True, help="Run standard updates")
@click.option("-l/-L", "--large/--no-large", is_flag=True, help="Use more RAM")
@click.option("-s", "--state", type=str, help="State to enter by default")
@click.option("-c", "--config", type=P, help="Config part to use for the device")
@click.option("-w", "--watch", is_flag=True, help="monitor the target's output after setup")
@click.option("-C", "--cross", help="path to mpy-cross")
@click.option("-M", "--mode", help="Mode add_on (filename, section)")
@click.option(
    "-m",
    "--main",
    type=click.Path(dir_okay=False, readable=True, exists=True),
    help="file to use as main_.py",
)
@catch_errors
async def setup_(ctx, run_section=None, mode=None, **kw):
    """
    Initial sync of MoaT code to a MicroPython device.

    MoaT must not currently run on the target.
    """
    from .setup import setup  # noqa: PLC0415

    ocfg = ctx.obj.cfg["micro"]

    cfg = ctx.obj.mcfg
    cfg.args.config = get_part(
        ocfg,
        cfg.args.config[:-1] / f"{cfg.args.config[-1]}_{mode}"
        if mode is not None
        else cfg.args.config,
    )

    default = {
        k: v
        for k, v in kw.items()
        if ctx.get_parameter_source(k) == click.core.ParameterSource.DEFAULT
    }
    param = {
        k: v
        for k, v in kw.items()
        if ctx.get_parameter_source(k) != click.core.ParameterSource.DEFAULT
    }

    # teach the 'large' flag to be ternary
    if "large" in default:
        default["large"] = None

    st = {
        k: (v if v != "-" else NotGiven) for k, v in cfg.setdefault("args", {}).items() if k in kw
    }

    st = combine_dict(param, st, default)

    st["config_file"] = f"moat_{mode}.cfg" if mode is not None else "moat.cfg"

    run = st["run"]
    if run:
        if run_section is not None:
            st["run"] = ctx.obj.cfg.micro._get(run_section)  # noqa: SLF001
        else:
            try:
                st["run"] = ctx.obj.cfg.micro._get(P("setup.run"))  # noqa: SLF001
            except KeyError:
                st["run"] = ctx.obj.cfg.micro["run"]

    return await setup(cfg, **st)


@cli.command("sync", short_help="Sync MoaT code")
@click.pass_context
@click.option(
    "-s",
    "--source",
    type=click.Path(dir_okay=True, file_okay=True, path_type=anyio.Path),
    multiple=True,
    help="more files to sync",
)
@click.option("-d", "--dest", type=str, default=".", help="Destination path")
@click.option("-C", "--cross", help="path to mpy-cross")
@click.option("-B/-b", "--boot/--no-boot", help="Reboot after updating")
@click.option("-U/-V", "--update/--no-update", is_flag=True, help="Run standard updates")
@catch_errors
async def sync_(ctx, **kw):
    """
    Sync of MoaT code to a running MicroPython device.

    """
    from .files import MoatFSPath  # noqa: PLC0415
    from .setup import do_copy, do_update  # noqa: PLC0415

    obj = ctx.obj
    cfg = obj.mcfg

    default = {
        k: v
        for k, v in kw.items()
        if ctx.get_parameter_source(k) == click.core.ParameterSource.DEFAULT
    }

    mydir = FSPath(__file__).parent.parent.parent
    mpydir = mydir / "ext" / "micropython"
    cross = mpydir / "mpy-cross" / "build" / "mpy-cross"
    if not cross.exists():
        cross = which("mpy-cross")
    default["cross"] = cross

    param = {
        k: v
        for k, v in kw.items()
        if ctx.get_parameter_source(k) != click.core.ParameterSource.DEFAULT
    }
    st = {k: (v if v != "-" else NotGiven) for k, v in cfg.get("sync", {}).items() if k in kw}
    st = combine_dict(param, st, default)

    async def syn(source=(), dest=".", cross=None, update=False, boot=False):
        if cross == "-":
            cross = None
        dest = dest.lstrip("/")  # needs to be relative
        if not update and not source:
            if obj.debug:
                print("Nothing to do.", file=sys.stderr)
            return
        if dest is None:
            raise click.UsageError("Destination cannot be empty")
        async with (
            RootCmd(cfg) as dsp,
            dsp.sub_at(cfg.remote) as cfr,
            cfr.sub_at(cfg.path.fs) as rfs,
            cfr.sub_at(cfg.path.sys) as rsys,
        ):
            root = MoatFSPath("/").connect_repl(rfs)
            dst = MoatFSPath(dest).connect_repl(rfs)

            async def hsh(p):
                return await rsys.hash(p=p)

            if update:
                await do_update(dst, root, cross, hsh)
            for s in source:
                await do_copy(s, root, dest, cross if cross != "-" else None)
            if boot:
                await rsys.boot(code="SysBooT", m=1)

    await syn(**st)
    if obj.debug:
        print(" " * 50, end="\r")


@cli.command(short_help="Reboot MoaT node")
@click.pass_obj
@click.option("-s", "--state", help="State after reboot")
@catch_errors
async def boot(obj, state):
    """
    Restart a MoaT node

    """
    cfg = obj.mcfg
    async with (
        RootCmd(cfg) as dsp,
        dsp.sub_at(cfg.remote) as cfr,
        cfr.sub_at(cfg.path.sys) as sd,
    ):
        if state:
            await sd.state(state=state)

        # reboot via the multiplexer
        logger.info("Rebooting target.")
        await sd.boot(code="SysBooT", m=1)

        # await t.send(["sys","boot"], code="SysBooT")
        await anyio.sleep(2)

        res = await sd("test")
        assert res == TEST_MAGIC, res

        res = await sd("ping", m="pong")
        if res != "R:pong":
            raise RuntimeError("wrong reply")
        print("Success:", res, file=sys.stderr)


@cli.command(short_help="Send a MoaT command")
@click.pass_obj
@click.argument("path", nargs=1, type=P)
@attr_args(with_path=True, with_proxy=True, with_arglist=True)
@click.option("-a", "--parts", is_flag=True, help="Retrieve a possibly-partial result")
@click.option("-t", "--time", is_flag=True, help="Time the command")
@click.option("-S", "--stream", is_flag=True, help="Get data stream from remote")
@catch_errors
async def cmd(obj, path, time, parts, stream, **attrs):
    """
    Send a MoaT command.

    The command is prefixed by the "micro.connect.remote" option;
    use "moat micro -R ‹path› cmd …" to change it if necessary.

    Positional arguments may follow the command name. They are parsed as
    with `--set`.
    """
    if parts and stream:
        raise click.UsageError("Parts and Stream modes don't like each other (yet).")
    cfg = obj.mcfg
    val = process_args({None: []}, no_path=True, **attrs)
    args = val.pop(None)
    logger.debug(
        "Command: %s %s %s",
        cfg.remote + path,
        "-" if not args else " ".join(str(a) for a in args),
        "-" if not val else " ".join(f"{k}={v!r}" for k, v in val.items()),
    )

    from time import monotonic as tm  # noqa: PLC0415

    from moat.util.times import humandelta  # noqa: PLC0415

    t1 = tm()
    async with (
        SigCancel(),
        RootCmd(cfg) as dsp,
        dsp.sub_at(cfg.remote) as cfr,
    ):
        try:
            t2 = tm()
            cmd = cfr.sub_at(path)
            if parts:
                from moat.lib.rpc import SubStore  # noqa: PLC0415

                res = await SubStore(cmd).get(*args, **val)
            elif stream:
                async with cmd.stream_in(*args, **val) as ms:
                    d = ms.to_list()
                    if d:
                        yprint(d)
                        print("---", file=obj.stdout)
                    async for m in ms:
                        yprint(m.to_list())
                        print("---", file=obj.stdout)
                    d = ms.to_list()
                    if d:
                        yprint(d)
                        print("---", file=obj.stdout)
            else:
                res = await cmd(*args, **val)
        except RemoteError as err:
            t3 = tm()
            yprint(dict(e=str(err.args[0])), stream=obj.stdout)
        else:
            if not stream:
                t3 = tm()
                if isinstance(res, Msg):
                    res = [res.args, res.kw]
                yprint(res, stream=obj.stdout)
    if time:
        print(f"{humandelta(t3 - t2)} (setup {humandelta(t2 - t1)})")


@cli.command(short_help="Read a console")
@click.pass_obj
@click.argument("path", nargs=1, type=P)
@catch_errors
async def cons(obj, path):
    """
    Read a Moat console.

    The command repeatedly calls the "crd" function on the given path and
    streams the result.
    """
    cfg = obj.mcfg
    async with (
        RootCmd(cfg) as dsp,
        dsp.sub_at(cfg.remote) as cfr,
    ):
        crd = cfr.sub_at(path).crd
        while True:
            try:
                res = await crd()
            except RemoteError as err:
                print(f"\nERR: {err!r}\n")
            else:
                print(res.decode("utf-8", errors="replace"), end="")


@cli.command("cfg", short_help="Get / Update the configuration")
@click.pass_obj
@click.option("-c", "--config", is_flag=True, help="Use MoaT config")
@click.option("-C", "--config-sat", is_flag=True, help="Use the satellite config")
@click.option("-r", "--read", type=click.File("r"), help="Read config from this file")
@click.option("-R", "--read-sat", help="Read config file from the satellite")
@click.option("-w", "--write", type=click.File("w"), help="Write config to this file")
@click.option("--stdout", is_flag=True, hidden=True)
@click.option("-W", "--write-sat", help="Write config file to the satellite")
@click.option("-S", "--sync", is_flag=True, help="Sync the satellite after writing")
@click.option("-D", "--send", is_flag=True, help="Also send to the satellite")
@click.option("-A", "--auth", is_flag=True, help="Authoritative: clear other satellite data")
@click.option(
    "-I",
    "--prio",
    type=str,
    default="mlrsa",
    help="Source priority: (first match wins)"
    " m:moat config l:local file r:remote file s:satellite config a:s-arguments",
)
@attr_args(with_proxy=True)
@catch_errors
async def cfg_(
    obj,
    read,
    read_sat,
    config,
    config_sat,
    write,
    write_sat,
    sync,
    send,
    stdout,
    auth,
    prio,
    **attrs,
):
    """
    Print, update and/or modify a remote configuration.

    The default for reading a config is the remote system. If modifiers (``-s …``)
    are used but no write option (``-w``, ``-W``), the remote config is
    updated in-place without being transferred.

    The default for writing is the remote system if modifiers have been
    used (and no write option). The remote config is updated anyway if you
    use ``--sync``.

    Configuration is read as YAML from stdin (``-r -``) or a
    file (``-r PATH``), as CBOR from Flash (``-R xx.cfg``), or from MoaT
    config (``-c``). Use satellite's memory (``-R -``),

    If you use multiple config sources, the ``--prio`` argument controls
    which data are preferred. The last match wins.

    An "app" section must be present if you write a complete configuration
    to the satellite.

    This command assumes that the remote system supplies a ``cfg.Cmd`` app at
    path "r.s.cfg_", and a ``fs.Cmd`` app at path "r.s.fs". You can change
    these paths with the ``-P fs ‹path›`` and ``-P cfg ‹path›`` options to
    ``moat micro``.
    """
    # add missing keys to prio
    for a in "mlrsa":
        if a not in prio:
            prio = a + prio

    if write == "-":
        stdout = True
        write = None

    has_attrs = any(a for a in attrs.values())

    arg_src = read or read_sat or config or config_sat
    arg_dest_ns = write or write_sat or stdout
    arg_dest = arg_dest_ns or sync

    if auth:
        sync = True
    if sync:
        send = True

    if (has_attrs or sync) and not auth and not arg_dest_ns:
        # we might be OK with incomplete data
        if not arg_dest_ns:
            send = True
        elif arg_dest_ns and not arg_src:
            # we have a destination but not complete source? that won't do
            config_sat = True
    else:
        # no attrs or auth = we want complete data
        if not arg_dest:
            # default to writing to stdout
            stdout = True
        if not arg_src:
            # default to input from remote config
            config_sat = True

    # do we need complete full data?
    if arg_dest_ns or auth:
        # do we not read it from anywhere?
        if not read and not read_sat and not config and not config_sat:
            # well, do read it then.
            config_sat = True

    mcfg = obj.mcfg

    if read_sat or write_sat:
        from .files import MoatFSPath  # noqa: PLC0415

    p_cfg = mcfg.path.get("cfg", P("cfg_"))
    p_fs = mcfg.path.get("fs", P("fs"))
    async with (
        SigCancel(),
        RootCmd(mcfg) as dsp,
        dsp.sub_at(mcfg.remote) as cfr,
        cfr.cfg_at(p_cfg) as cf,
        cfr.sub_at(p_fs) as fs,
    ):
        codec = get_codec("std-cbor")

        rcfg = attrdict()
        for k in prio:
            if k == "a":
                rcfg = process_args(rcfg, **attrs)
            elif k == "m":
                if config:
                    mc = obj.cfg.micro
                    merge(rcfg, mc.get_(mc.setup.args.config), replace=True)
            elif k == "l":
                if read:
                    merge(rcfg, yload(read if read != "-" else sys.stdin), replace=True)
            elif k == "r":
                if read_sat:
                    p = MoatFSPath(read_sat).connect_repl(fs)
                    d = await p.read_bytes(chunk=64)
                    merge(rcfg, codec.decode(d), replace=True)
            elif k == "s":
                if config_sat:
                    merge(rcfg, await cf.get(), replace=True)
            else:
                raise click.UsageError(f"Unknown prio key: {k!r}")

        if write_sat:
            if "app" in rcfg:
                p = MoatFSPath(write_sat).connect_repl(fs)
                d = codec.encode(rcfg)
                await p.write_bytes(d, chunk=64)
            else:
                print("No 'app' section. Not writing.", file=sys.stderr)

        if send:
            if not auth or ("app" in rcfg and "app" in rcfg.app):
                await cf.set(rcfg, sync=sync, replace=auth)
            else:
                print("No 'app' section. Not replacing.", file=sys.stderr)

        if write:
            yprint(rcfg, stream=write)

        if stdout:
            yprint(rcfg, stream=obj.stdout)


@cli.command("run", short_help="Run the multiplexer")
@click.pass_obj
@catch_errors
async def run_(obj):
    """
    Run the MoaT stack.
    """
    async with SigCancel(), RootCmd(obj.mcfg):
        await idle()


@cli.command("mount")
@click.option("-b", "--blocksize", type=int, help="Max read/write message size", default=256)
@click.argument("path", type=click.Path(file_okay=False, dir_okay=True), nargs=1)
@click.pass_obj
@catch_errors
async def mount_(obj, path, blocksize):
    """Mount a controller's file system on the host"""
    from moat.micro.fuse import wrap  # noqa: PLC0415

    cfg = obj.mcfg

    async with (
        SigCancel(),
        RootCmd(cfg) as dsp,
        dsp.sub_at(cfg.remote) as cfr,
        cfr.sub_at(cfg.path.fs) as sd,
        wrap(sd, path, blocksize=blocksize, debug=max(obj.debug - 1, 0)),
    ):
        if obj.debug:
            print("Mounted.")
        await idle()


@cli.command("rom")
@click.option("-d", "--device", type=int, help="ROMFS segment to use", default=0)
@click.argument("path", type=click.Path(file_okay=False, dir_okay=True), nargs=1)
@click.pass_obj
@catch_errors
async def rom(obj, path, device):
    """Send a file system to the device's ROM"""
    path  # noqa:B018
    raise NotImplementedError

    cfg = obj.mcfg

    async with (
        RootCmd(cfg) as dsp,
        dsp.sub_at(cfg.remote) as cfr,
        cfr.sub_at(cfg.path.rom) as sd,
    ):
        res = await sd.n()
        if device >= res:
            if res == 1:
                raise RuntimeError("Device has only one ROMFS.")
            elif not res:
                raise RuntimeError("Device does not have ROMFS.")
            raise RuntimeError("Device 0…{res-1} only.")

        _nblk, _blksz = await sd.stat()

        if obj.debug:
            print("Building ROMFS.")

        # TODO


@cli.command("path")
@click.pass_obj
@click.option("-m", "--manifest", is_flag=True, help="main manifest")
@catch_errors
async def path_(obj, manifest):
    """Path to the embedded system's files"""

    import pathlib  # noqa: PLC0415

    if manifest:
        import moat.micro._embed._tag as m  # noqa: PLC0415

        print(m.__file__.replace("_tag", "manifest"), file=obj.stdout)
        return

    import moat.micro._embed  # noqa: PLC0415

    for p in moat.micro._embed.__path__:  # noqa:SLF001
        p = pathlib.Path(p) / "lib"  # noqa:PLW2901
        if p.exists():
            print(p)


@cli.command
@click.pass_obj
async def repl(obj):
    """Connect to the Python prompt on a remote"""

    cfg = obj.mcfg
    from moat.lib.stream import FilenoBuf  # noqa:PLC0415

    async with (
        SigCancel(),
        RootCmd(cfg) as dsp,
        dsp.sub_at(cfg.remote) as cfr,
        cfr.sub_at(cfg.terminal) as cft,
        cft().stream() as t1,
        FilenoBuf({}, 0, 1) as t2,
        anyio.create_task_group() as tg,
    ):
        print("--- Connected.  Quit: Ctrl+] --- ", file=sys.stderr)

        @tg.start_soon
        async def t_12():
            async for msg in t1:
                tx = msg[0]
                if tx and tx[0] == 0:
                    await t2.wr("[………]".encode("utf-8"))
                    tx = tx[1:]
                await t2.wr(tx)
            tg.cancel_scope.cancel()

        buf = bytearray(80)
        while True:
            try:
                n = await t2.rd(buf)
            except EOFError:
                tg.cancel_scope.cancel()
                break
            if b"\x1d" in buf[:n]:
                tg.cancel_scope.cancel()
                break
            await t1.send(buf[:n].replace(b"\n", b"\r"))
    print("\nTerminated.", file=sys.stderr)



"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.
"""

from __future__ import annotations

import struct

from moat.util import OutOfData

from ._base import Codec as _Codec
from ._base import NoCodecError

# Typing
from typing import TYPE_CHECKING, cast  # isort:skip

try:
    from micropython import const  # type: ignore[import-not-found]  # MicroPython only
except ImportError:

    def const(x: int) -> int:
        return x


if TYPE_CHECKING:
    from ._base import ByteType, VarByteType

    from typing import Any


__all__ = ["CBOR_TAG_CBOR_FILEHEADER", "CBOR_TAG_CBOR_LEADER", "Codec", "ExtraData", "Tag"]


CBOR_TAG_CBOR_FILEHEADER = const(55799)  # single CBOR content
CBOR_TAG_CBOR_LEADER = const(55800)  # header for multiple CBOR items


class Tag:
    "a random CBOR tag"

    def __init__(self, tag: int, value: Any = None):
        self.tag = tag
        self.value = value

    def __repr__(self):
        return f"Tag({self.tag!r}, {self.value!r})"

    def __eq__(self, other):
        if not isinstance(other, Tag):
            return False
        return self.tag == other.tag and self.value == other.value

    def __hash__(self):
        return hash(self.tag) ^ hash(self.value)


# Original Copyright 2014-2015 Brian Olson
# Apache 2.0 license
# http://docs.ros.org/en/noetic/api/rosbridge_library/html/cbor_8py_source.html
# rather heavily modified

CBOR_TYPE_MASK = const(0xE0)  # top 3 bits
CBOR_INFO_BITS = const(0x1F)  # low 5 bits

CBOR_UINT = const(0x00)
CBOR_NEGINT = const(0x20)
CBOR_BYTES = const(0x40)
CBOR_TEXT = const(0x60)
CBOR_ARRAY = const(0x80)
CBOR_MAP = const(0xA0)
CBOR_TAG = const(0xC0)
CBOR_7 = const(0xE0)  # float and other types

CBOR_UINT8_FOLLOWS = const(24)  # 0x18
CBOR_UINT16_FOLLOWS = const(25)  # 0x19
CBOR_UINT32_FOLLOWS = const(26)  # 0x1a
CBOR_UINT64_FOLLOWS = const(27)  # 0x1b
CBOR_VAR_FOLLOWS = const(31)  # 0x1f

CBOR_BREAK = const(0xFF)

CBOR_FALSE = const(CBOR_7 | 20)
CBOR_TRUE = const(CBOR_7 | 21)
CBOR_NULL = const(CBOR_7 | 22)
CBOR_UNDEFINED = const(CBOR_7 | 23)  # js 'undefined' value

CBOR_FLOAT16 = const(CBOR_7 | 25)
CBOR_FLOAT32 = const(CBOR_7 | 26)
CBOR_FLOAT64 = const(CBOR_7 | 27)

CBOR_TAG_BIGNUM = const(2)  # big endian byte string follows
CBOR_TAG_NEGBIGNUM = const(3)  # big endian byte string follows

_CBOR_TAG_BIGNUM_BYTES = struct.pack("B", CBOR_TAG | CBOR_TAG_BIGNUM)
_CBOR_TAG_NEGBIGNUM_BYTES = struct.pack("B", CBOR_TAG | CBOR_TAG_NEGBIGNUM)


def _bignum_to_bytearray(val):
    # separate because µPython needs to do this differently
    return val.to_bytes((val.bit_length() + 7) // 8, "big")


class ExtraData(ValueError):
    "buffer contains data beyond end of encoded object"

    def __str__(self):
        return "decoder received extra data."


class Codec(_Codec):
    "Basic CBOR codec"

    _buffer: bytes | bytearray = b""
    _buf_pos: int = 0

    _buf_out: bytearray | None = None

    def __init__(self, use_attrdict: bool = False, **kw):
        super().__init__(**kw)
        self.use_attrdict = use_attrdict
        self.__kw = kw

    def copy(self) -> Codec:
        "copy me"
        return Codec(use_attrdict=self.use_attrdict, **self.__kw)

    def encode(self, obj: Any) -> ByteType:
        """
        Pack @obj, return the resulting bytes.
        """
        if self._buf_out is not None:
            raise RuntimeError("Codec is busy")

        self._buf_out = bytearray()
        try:
            self._enc_any(obj)
            return self._buf_out
        finally:
            self._buf_out = None  # always reset

    def decode(self, data: ByteType) -> Any:
        """
        Unpack @data, return the resulting object.
        """
        if self._buffer:
            raise RuntimeError("Codec is busy")

        self._buffer = cast(bytes, data)
        try:
            res = self._dec_any()
            # chop off the part we've read
            if self._buf_pos != len(data):
                raise ExtraData
            return res
        finally:
            self._buffer = b""  # always reset
            self._buf_pos = 0

    def feed(self, data: ByteType) -> None:
        "Add additional input"
        if self._buffer and self._buf_pos < len(self._buffer):
            if self._buf_pos == 0:
                self._buffer = cast(bytes, self._buffer + data)
                return
            data = cast(bytes, self._buffer[self._buf_pos :] + data)
        if isinstance(data, memoryview):
            data = bytearray(data)
        self._buffer = cast(bytes, data)
        self._buf_pos = 0

    def unfeed(self, buf: VarByteType) -> int:
        "take from the decoder's buffer"
        if not self._buffer:
            return 0
        if buf is None:
            if isinstance(self._buffer, memoryview):
                self._buffer = bytearray(self.buffer)
            return 0
        i = self._buf_pos
        n = min(len(buf), len(self._buffer) - i)
        if n <= 0:
            return 0
        i_n = i + n
        buf[:n] = self._buffer[i:i_n]
        self._buf_pos = i_n
        return n

    def _enc_int(self, val):
        "return bytes representing int val in CBOR"
        if val < 0:
            val = -1 - val
            return self._enc_type_num(CBOR_NEGINT, val)
        return self._enc_type_num(CBOR_UINT, val)

    def _enc_float(self, val):
        w = self._w
        try:
            ff = struct.pack("!f", val)
        except OverflowError:
            pass
        else:
            if struct.unpack("!f", ff)[0] == val:  # no loss
                try:
                    fe = struct.pack("!e", val)
                except OverflowError:
                    pass
                else:
                    if struct.unpack("!e", fe)[0] == val:  # no loss either
                        return w(struct.pack("!B", CBOR_FLOAT16) + fe)
                return w(struct.pack("!B", CBOR_FLOAT32) + ff)
        return w(struct.pack("!Bd", CBOR_FLOAT64, val))

    def _enc_type_num(self, cbor_type, val) -> None:
        """
        For some CBOR primary type [0..7] and an auxiliary unsigned
        number, return CBOR encoded bytes
        """
        w = self._w

        if val is None:
            return w(struct.pack("B", cbor_type | CBOR_VAR_FOLLOWS))

        assert val >= 0
        if val <= 23:
            return w(struct.pack("B", cbor_type | val))
        if val <= 0x0FF:
            return w(struct.pack("BB", cbor_type | CBOR_UINT8_FOLLOWS, val))
        if val <= 0x0FFFF:
            return w(struct.pack("!BH", cbor_type | CBOR_UINT16_FOLLOWS, val))
        if val <= 0x0FFFFFFFF:
            return w(struct.pack("!BI", cbor_type | CBOR_UINT32_FOLLOWS, val))
        if val <= 0x0FFFFFFFFFFFFFFFF:
            return w(struct.pack("!BQ", cbor_type | CBOR_UINT64_FOLLOWS, val))

        # too large: write a bignum
        if cbor_type == CBOR_UINT:
            w(_CBOR_TAG_BIGNUM_BYTES)
        elif cbor_type == CBOR_NEGINT:
            w(_CBOR_TAG_NEGBIGNUM_BYTES)
        else:
            raise ValueError("value too big: " + repr(val))
        outb = _bignum_to_bytearray(val)
        self._enc_type_num(CBOR_BYTES, len(outb))
        w(outb)

    def _enc_string(self, val):
        if isinstance(val, str):
            val = val.encode("utf8")
            self._enc_type_num(CBOR_TEXT, len(val))
        else:
            self._enc_type_num(CBOR_BYTES, len(val))
        self._w(val)

    def _enc_array(self, arr):
        try:
            self._enc_type_num(CBOR_ARRAY, len(arr))
        except TypeError:
            self._enc_type_num(CBOR_ARRAY, None)
            for x in arr:
                self._enc_any(x)
            self._w(struct.pack("B", CBOR_BREAK))
        else:
            for x in arr:
                self._enc_any(x)

    def _enc_dict(self, d):
        self._enc_type_num(CBOR_MAP, len(d))
        for k, v in d.items():
            self._enc_any(k)
            self._enc_any(v)

    def _enc_bool(self, b):
        self._w(struct.pack("B", CBOR_TRUE if b else CBOR_FALSE))

    def _enc_tag(self, t, val=...):
        if val is Ellipsis:
            t, val = t.tag, t.value
        self._enc_type_num(CBOR_TAG, t)
        self._enc_any(val)

    def _w(self, d):
        assert self._buf_out is not None  # set by encode() before calling _enc_any()
        self._buf_out.extend(d)

    def _enc_any(self, ob):
        w = self._w
        if ob is None:
            w(struct.pack("B", CBOR_NULL))
        elif ob is Ellipsis:
            w(struct.pack("B", CBOR_UNDEFINED))
        elif isinstance(ob, bool):
            self._enc_bool(ob)
        elif isinstance(ob, (str, bytes, bytearray, memoryview)):
            self._enc_string(ob)
        elif isinstance(ob, (list, tuple)):
            self._enc_array(ob)
        # TODO: accept other enumerables and emit a variable length array
        elif isinstance(ob, dict):
            self._enc_dict(ob)
        elif isinstance(ob, float):
            self._enc_float(ob)
        elif isinstance(ob, int):
            self._enc_int(ob)
        elif isinstance(ob, Tag):
            self._enc_tag(ob)
        else:
            self._enc_tag(*self.ext.encode(self, ob))  # pyright:ignore

    # Decoder

    def _read_byte(self):
        return self._read(1)[0]

    def __iter__(self):
        return self

    def __next__(self):
        i = self._buf_pos
        try:
            return self._dec_any()
        except OutOfData:
            self._buf_pos = i
            raise StopIteration from None
        except BaseException:
            self._buf_pos = i
            raise
        finally:
            if len(self._buffer) == self._buf_pos:
                self._buffer = b""
                self._buf_pos = 0

    def _dec_tag_aux(self, tb):
        tag = tb & CBOR_TYPE_MASK
        tag_aux = tb & CBOR_INFO_BITS
        if tag_aux <= 23:
            aux = tag_aux
        elif tag_aux == CBOR_UINT8_FOLLOWS:
            data = self._read(1)
            aux = struct.unpack_from("!B", data, 0)[0]
        elif tag_aux == CBOR_UINT16_FOLLOWS:
            data = self._read(2)
            aux = struct.unpack_from("!H", data, 0)[0]
        elif tag_aux == CBOR_UINT32_FOLLOWS:
            data = self._read(4)
            aux = struct.unpack_from("!I", data, 0)[0]
        elif tag_aux == CBOR_UINT64_FOLLOWS:
            data = self._read(8)
            aux = struct.unpack_from("!Q", data, 0)[0]
        else:
            if tag_aux != CBOR_VAR_FOLLOWS:
                raise ValueError(f"bogus tag {tb:02x}")
            aux = None

        return tag, aux

    def _read(self, n):
        # (int) -> bytearray
        if (nb := len(self._buffer) - self._buf_pos) < n:
            raise OutOfData(n - nb)
        i = self._buf_pos
        self._buf_pos += n
        return self._buffer[i : i + n]

    def _dec_var_array(self):
        ob = []
        tb = self._read_byte()
        while tb != CBOR_BREAK:
            subob = self._dec_tagged(tb)
            ob.append(subob)
            tb = self._read_byte()
        return ob

    def _dec_var_map(self):
        ob = {}
        tb = self._read_byte()
        while tb != CBOR_BREAK:
            subk = self._dec_tagged(tb)
            subv = self._dec_any()
            ob[subk] = subv
            tb = self._read_byte()
        return ob

    def _dec_array(self, aux):
        ob = []
        for _ in range(aux):
            subob = self._dec_any()
            ob.append(subob)
        return ob

    def _dec_map(self, aux):
        ob = {}
        for _ in range(aux):
            subk = self._dec_any()
            subv = self._dec_any()
            ob[subk] = subv
        return ob

    def _dec_any(self):
        return self._dec_tagged(self._read_byte())

    def _dec_tagged(self, tb):
        # Some special cases of CBOR_7 best handled by special struct.unpack logic here
        if tb == CBOR_FLOAT16:
            data = self._read(2)
            pf = struct.unpack_from("!e", data, 0)
            return pf[0]
        elif tb == CBOR_FLOAT32:
            data = self._read(4)
            pf = struct.unpack_from("!f", data, 0)
            return pf[0]
        elif tb == CBOR_FLOAT64:
            data = self._read(8)
            pf = struct.unpack_from("!d", data, 0)
            return pf[0]

        tag, aux = self._dec_tag_aux(tb)

        if tag == CBOR_UINT:
            return aux
        elif tag == CBOR_NEGINT:
            return -1 - cast(int, aux)
        elif tag == CBOR_BYTES:
            return self._dec_bytes(aux, CBOR_BYTES)
        elif tag == CBOR_TEXT:
            return str(self._dec_bytes(aux, CBOR_TEXT), "utf8")
        elif tag == CBOR_ARRAY:
            if aux is None:
                return self._dec_var_array()
            return self._dec_array(aux)
        elif tag == CBOR_MAP:
            if aux is None:
                return self._dec_var_map()
            return self._dec_map(aux)
        elif tag == CBOR_TAG:
            ob = self._dec_any()
            aux = cast(int, aux)
            try:
                return self.ext.decode(self, aux, ob)
            except NoCodecError:
                return Tag(aux, ob)
        else:
            assert tag == CBOR_7
            if tb == CBOR_TRUE:
                return True
            if tb == CBOR_FALSE:
                return False
            if tb == CBOR_NULL:
                return None
            if tb == CBOR_UNDEFINED:
                return Ellipsis
            raise ValueError(f"unknown cbor tag 7 byte: {tb:02x}")

    def _dec_bytes(self, aux, btag):
        # TODO: limit to some maximum number of chunks and some maximum total bytes
        if aux is not None:
            # simple case
            ob = self._read(aux)
            return ob
        # read chunks
        chunklist = []
        while True:
            tb = self._read_byte()
            if tb == CBOR_BREAK:
                break
            tag, aux = self._dec_tag_aux(tb)
            if tag != btag:
                raise ValueError(f"var length with unexpected component:{tag:02x} {btag:02x}")
            ob = self._read(aux)
            chunklist.append(ob)
        return b"".join(chunklist)



"""
Module for pins
"""

from __future__ import annotations

import machine as M

from moat.compat import AC_use, Event, TaskGroup, sleep_ms, to_thread
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg


class _ADC(M.ADC):
    def __new__(cls, cfg, **kw):
        kw["id"] = M.Pin(cfg["pin"])
        self = super().__new__(**kw)
        return self

    def __init__(self, cfg, **kw):  # noqa:ARG002
        self.t = cfg.get("t", 1000)
        self.nn = cfg.get("nn", 1)
        self.dly = cfg.get("delay", 1)
        self.factor = cfg.get("factor", 1) / self.n / self.nn
        self.offset = cfg.get("offset", 0)

        self.val = 0
        self.evt = Event()
        self.delta = cfg.get("delta", 3)

    async def scan(self):
        self.val = await self.read()
        self.evt.set()
        self.evt = Event()

        while True:
            await sleep_ms(self.t)
            v = await self.read()
            ov, self.val = self.val, v
            if abs(ov - v) >= self.delta:
                self.evt.set()
                self.evt = Event()

    async def read(self):
        c = 0
        for a in range(self.nn):
            if a:
                await sleep_ms(self.dly)
            for _ in range(self.n):
                c += await to_thread(self.read_u16)
        res = c * self.factor + self.offset
        return res

    async def __aenter__(self):
        return self

    async def __aexit__(self, *err):
        pass

    def __aiter__(self):
        return self

    async def __anext__(self):
        await self.evt.wait()
        return self.val


class ADC(BaseCmd):
    """
    This is a basic analog input.

    Iterating it yields a new value whenever the value changes by a given
    threshold.
    """

    doc = dict(
        _c=dict(
            _d="analog input",
            td="int:pin",
            ts="int:sample timer (ms)",
            delta="int:abs difference",
            factor="float:factor (1)",
            offset="float:offset (0)",
            delay="int:avg delay (1,ms)",
            nn="int:avg count",
        ),
        _d="read",
        o="any:wait for val to not be this",
        d="int:delta",
        _s=[dict(_r="float:value")],
        _o="float:value",
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        kw = {}
        if "ts" in cfg:
            kw["sample_ns"] = int(cfg["ts"] * 1000000)
        if "atten" in cfg:
            kw["atten"] = cfg["atten"]
        self.adc = _ADC(cfg, **kw)

    async def setup(self):
        "initialization, triggers change"
        await super().setup()
        if getattr(self, "tg", None) is None:
            self.tg = await AC_use(self, TaskGroup())
        await self.tg.spawn(self.pin.scan)

    async def stream(self, msg: Msg):
        "read. Wait for changes if @o (old value) is not None"
        o = msg.get("o", None)
        d = msg.get("d", 0)
        if msg.can_stream:
            while True:
                v = self.adc.val
                if abs(v - o) < d:
                    await self.adc.evt.wait()
                    continue
                await msg.result(v)
                o = v
                await self.adc.evt.wait()
        else:
            while abs(self.adc.val - o) < d:
                await self.adc.evt.wait()
            await msg.result(self.adc.val)
